import copy
from datetime import date, timedelta
from typing import Dict, Tuple

import pandas as pd
from pytz import timezone
from tqdm import tqdm

from liualgotrader.common import config
from liualgotrader.common.data_loader import DataLoader  # type: ignore
from liualgotrader.common.database import fetch_as_dataframe
from liualgotrader.common.tlog import tlog
from liualgotrader.common.types import AssetType
from liualgotrader.models.accounts import Accounts
from liualgotrader.models.optimizer import OptimizerRun
from liualgotrader.models.portfolio import Portfolio
from liualgotrader.trading.trader_factory import trader_factory

est = timezone("America/New_York")


async def load_trades_for_period(
    env: str, from_date: date, to_date: date
) -> pd.DataFrame:
    query = f"""
    SELECT client_time, symbol, operation, qty, price, algo_name
    FROM 
        new_trades as t, algo_run as a
    WHERE 
        t.algo_run_id = a.algo_run_id AND 
        t.tstamp >= '{from_date}' AND 
        t.tstamp < '{to_date}' AND
        t.expire_tstamp is null AND
        a.algo_env = '{env}'
    ORDER BY symbol, tstamp
    """
    return await fetch_as_dataframe(query)


async def portfolio_return(
    env: str, start_date: date
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    df = await load_trades_for_period(
        env, start_date, date.today() + timedelta(days=1)
    )

    # day X strategy = total return
    table: Dict = {}
    invested_table: Dict = {}
    percentages_table: Dict = {}
    for _, row in df.iterrows():
        d = pd.Timestamp(pd.to_datetime(row["client_time"]).date(), tz=est)
        if d not in table:
            table[d] = {}
            invested_table[d] = {}
            percentages_table[d] = {}
            table[d]["revenue"] = 0.0
            invested_table[d]["invested"] = 0.0

        strat = row["algo_name"]
        if strat not in table[d]:
            table[d][strat] = 0.0
            invested_table[d][strat] = 0.0
            percentages_table[d][strat] = 0.0
        delta = (
            (1.0 if row["operation"] == "sell" and row["qty"] > 0 else -1.0)
            * float(row["price"])
            * row["qty"]
        )
        table[d][strat] += delta
        table[d]["revenue"] += delta
        if row["operation"] == "buy":
            invested_table[d][strat] += row["qty"] * float(row["price"])
            invested_table[d]["invested"] += row["qty"] * float(row["price"])
        percentages_table[d][strat] = (
            (100.0 * table[d][strat] / invested_table[d][strat])
            if invested_table[d][strat] != 0
            else 0.0
        )
        percentages_table[d]["revenue"] = (
            (100.0 * table[d]["revenue"] / invested_table[d]["invested"])
            if invested_table[d][strat] != 0
            else 0.0
        )

    return (
        pd.DataFrame.from_dict(table, orient="index").sort_index(),
        pd.DataFrame.from_dict(invested_table, orient="index").sort_index(),
        pd.DataFrame.from_dict(percentages_table, orient="index").sort_index(),
    )


async def load_trades(day: date, end_date: date = None) -> pd.DataFrame:
    query = f"""\x1f    SELECT t.*, a.batch_id, a.algo_name\x1f    FROM \x1f    new_trades as t, algo_run as a\x1f    WHERE \x1f        t.algo_run_id = a.algo_run_id AND \x1f        t.tstamp >= '{day}' AND \x1f        t.tstamp < '{end_date or day + timedelta(days=1)}' AND\x1f        t.expire_tstamp is null \x1f    ORDER BY symbol, tstamp\x1f    """
    return await fetch_as_dataframe(query)


async def load_trades_by_batch_id(batch_id: str) -> pd.DataFrame:
    query = f"""
        SELECT 
            t.*, a.batch_id, a.start_time, a.algo_name
        FROM 
            new_trades as t, algo_run as a
        WHERE 
            t.algo_run_id = a.algo_run_id AND 
            a.batch_id = '{batch_id}' AND
            t.expire_tstamp is null AND 
            price != 0.0 
        ORDER BY symbol, tstamp
    """
    df: pd.DataFrame = await fetch_as_dataframe(query)
    try:
        if not df.empty:
            df["client_time"] = pd.to_datetime(df["client_time"])
    except Exception:
        tlog(
            f"[Error] load_trades_by_batch_id({batch_id}) can't convert 'client_time' column to datetime"
        )
    return df


async def load_trades_by_portfolio(portfolio_id: str) -> pd.DataFrame:
    query = f"""
        SELECT 
            t.*, a.batch_id, a.start_time, a.algo_name
        FROM 
            new_trades as t, algo_run as a, portfolio_batch_ids as p
        WHERE 
            t.algo_run_id = a.algo_run_id AND 
            a.batch_id = p.batch_id AND
            p.portfolio_id = '{portfolio_id}' AND
            price != 0.0 AND 
            target_price is not null AND
            t.expire_tstamp is null 
        ORDER BY symbol, tstamp
    """
    df: pd.DataFrame = await fetch_as_dataframe(query)
    try:
        if not df.empty:
            df["client_time"] = pd.to_datetime(df["client_time"])
    except Exception:
        tlog(
            f"[Error] load_trades_by_portfolio({portfolio_id}) can't convert 'client_time' column to datetime"
        )
    return df


async def load_runs(day: date, end_date: date = None) -> pd.DataFrame:
    query = f"""\x1f     SELECT * \x1f     FROM \x1f     algo_run as t\x1f     WHERE \x1f         start_time >= '{day}' AND \x1f         start_time < '{end_date or day + timedelta(days=1)}'\x1f     ORDER BY start_time\x1f     """

    df = await fetch_as_dataframe(query)
    df.set_index("algo_run_id", inplace=True)
    return df


async def load_batch_list(day: date, env: str) -> pd.DataFrame:
    query = f"""
        SELECT DISTINCT a.batch_id, a.start_time
        FROM 
        new_trades as t, algo_run as a
        WHERE 
            t.algo_run_id = a.algo_run_id AND 
            t.tstamp >= '{day}' AND 
            t.tstamp < '{day + timedelta(days=1)}'AND
            t.expire_tstamp is null 
        """
    return await fetch_as_dataframe(query)


async def load_traded_symbols(batch_id: str) -> pd.DataFrame:
    query = f"""
        SELECT 
            DISTINCT t.symbol
        FROM 
            new_trades as t, algo_run as a
        WHERE 
            t.algo_run_id = a.algo_run_id AND 
            a.batch_id = '{batch_id}'
    """
    return await fetch_as_dataframe(query)


def calc_batch_revenue(
    symbol: str, trades: pd.DataFrame, batch_id: str = None
) -> float:
    symbol_df = trades[
        (trades["symbol"] == symbol)
        & (not batch_id or trades["batch_id"] == batch_id)
    ]
    rc = sum(
        (
            row["price"]
            * row["qty"]
            * (1 if row["operation"] == "sell" and row["qty"] > 0 else -1)
        )
        for index, row in symbol_df.iterrows()
    )
    return round(rc, 2)


def calc_revenue(symbol: str, trades: pd.DataFrame, env) -> float:
    symbol_df = trades[
        (trades["symbol"] == symbol) & (trades["algo_env"] == env)
    ]
    rc = sum(
        (
            row["price"]
            * row["qty"]
            * (1 if row["operation"] == "sell" and row["qty"] > 0 else -1)
        )
        for index, row in symbol_df.iterrows()
    )
    return round(rc, 2)


def count_trades(symbol, trades: pd.DataFrame, batch_id: str) -> int:
    symbol_df = trades[
        (trades["symbol"] == symbol) & (trades["batch_id"] == batch_id)
    ]
    return len(symbol_df.index)


def trades_analysis(trades: pd.DataFrame, batch_id: str) -> pd.DataFrame:
    min(trades["client_time"].tolist())
    trades_analytics = pd.DataFrame()
    trades["client_time"] = pd.to_datetime(trades["client_time"], utc=True)
    trades_analytics["symbol"] = trades.symbol.unique()
    trades_analytics["revenues"] = trades_analytics["symbol"].apply(
        lambda x: calc_batch_revenue(x, trades, batch_id)
    )
    trades_analytics["count"] = trades_analytics["symbol"].apply(
        lambda x: count_trades(x, trades, batch_id)
    )

    return trades_analytics


def symbol_trade_analytics(
    symbol_df: pd.DataFrame, plt
) -> Tuple[pd.DataFrame, float]:

    d = copy.deepcopy(symbol_df).drop(
        columns=[
            "trade_id",
            "algo_run_id",
            "symbol",
            "tstamp",
            "expire_tstamp",
            "batch_id",
            "start_time",
        ]
    )

    d["trade"] = symbol_df.apply(
        lambda row: row.price
        * row.qty
        * (1 if row.operation == "sell" and row.qty > 0 else -1),
        axis=1,
    )
    d["balance"] = d.trade.cumsum()

    for _, row in symbol_df.iterrows():
        plt.scatter(
            row["client_time"],
            row["price"],
            c="g" if row["operation"] == "buy" else "r",
            s=100,
        )

    return d, d.trade.sum()


def calc_symbol_trades_returns(
    symbol: str,
    symbol_trades: pd.DataFrame,
    daily_returns: pd.DataFrame,
    data_loader: DataLoader,
):
    t1 = t2 = None
    qty: float = 0.0
    eastern = timezone(
        "US/Eastern",
    )

    symbol_trades["qty"] = symbol_trades["qty"].astype(float)
    for _, row in symbol_trades.iterrows():
        if t1 is None:
            t1 = daily_returns.index.get_loc(str(row.client_time.date()))
        else:
            t2 = daily_returns.index.get_loc(str(row.client_time.date()))

        if t1 is not None and t2 is not None:
            for i in range(t1, t2):
                try:
                    value = (
                        qty
                        * data_loader[symbol].close[
                            eastern.localize(
                                daily_returns.index[i].to_pydatetime()
                            ).replace(
                                hour=9, minute=30, second=0, microsecond=0
                            )
                        ]
                    )
                    daily_returns.loc[
                        daily_returns.index[i],
                        "equity",
                    ] += value
                except Exception as e:
                    tlog(f"Error during calc_symbol_trades_returns(): {e}")

        if row.operation == "buy":
            qty += row.qty
        elif row.operation == "sell":
            qty -= row.qty

        if qty == 0:
            t1 = t2 = None
        elif t2 is not None:
            t1 = t2
            t2 = None

    if qty and t1 is not None:
        for i in range(t1, len(daily_returns.index)):
            try:
                value = (
                    qty
                    * data_loader[symbol].close[
                        eastern.localize(
                            daily_returns.index[i].to_pydatetime()
                        ).replace(hour=9, minute=30, second=0, microsecond=0)
                    ]
                )
            except ValueError:
                print("qty:", qty, t1)
                raise
            daily_returns.loc[daily_returns.index[i], "equity"] += value


def calc_symbol_state(
    symbol: str,
    symbol_trades: pd.DataFrame,
    data_loader: DataLoader,
) -> Tuple[float, float]:
    qty: float = 0
    for _, row in symbol_trades.iterrows():
        if row.operation == "buy":
            qty += row.qty
        elif row.operation == "sell":
            qty -= row.qty

    return qty, data_loader[symbol].close[-1]


async def calc_batch_returns(batch_id: str) -> pd.DataFrame:
    portfolio = await Portfolio.load_by_batch_id(batch_id)
    data_loader = DataLoader()
    trades = await load_trades_by_batch_id(batch_id)

    if trades.empty:
        return pd.DataFrame()
    start_date = trades.client_time.min().date()
    end_date = trades.client_time.max().date()
    trader = trader_factory()

    td = trader.get_trading_days(start_date=start_date, end_date=end_date)

    td["equity"] = 0.0
    td["cash"] = portfolio.portfolio_size

    num_symbols = len(trades.symbol.unique().tolist())
    for c, symbol in enumerate(trades.symbol.unique().tolist(), start=1):
        print(f"{symbol} ({c}/{num_symbols})")
        symbol_trades = trades[trades.symbol == symbol].sort_values(
            by="client_time"
        )
        calc_symbol_trades_returns(symbol, symbol_trades, td, data_loader)

    td["totals"] = td["equity"] + td["cash"]
    # td["date"] = pd.to_datetime(td.index)
    # td = td.set_index("date")
    return pd.DataFrame(td, columns=["equity", "cash", "totals"])


async def compare_to_symbol_returns(
    portfolio_id: str, symbol: str
) -> pd.DataFrame:
    data_loader = DataLoader()
    trades = await load_trades_by_portfolio(portfolio_id)
    start_date = trades.client_time.min().date()
    end_date = trades.client_time.max().date()
    trader = trader_factory()

    td = trader.get_trading_days(start_date=start_date, end_date=end_date)
    td[symbol] = td.apply(
        lambda row: data_loader[symbol].close[
            row.name.to_pydatetime().replace(tzinfo=est)
        ],
        axis=1,
    )
    return td[symbol]


async def get_cash(account_id: int, initial_cash: float) -> pd.DataFrame:
    df = await Accounts.get_transactions(account_id)
    df = df.groupby(df.index.date).sum()
    df.iloc[0].amount += initial_cash
    r = pd.date_range(start=df.index.min(), end=df.index.max(), name="date")
    df.reindex(r).fillna(method="ffill")
    df["cash"] = df.amount.cumsum()
    df.index.rename("date", inplace=True)
    return df


async def calc_portfolio_returns(portfolio_id: str) -> pd.DataFrame:
    portfolio = await Portfolio.load_by_portfolio_id(portfolio_id)

    data_loader = DataLoader()
    trades = await load_trades_by_portfolio(portfolio_id)

    if trades.empty:
        return pd.DataFrame()
    start_date = trades.client_time.min().date()
    end_date = trades.client_time.max().date()
    trader = trader_factory()

    if portfolio.asset_type == AssetType.US_EQUITIES:
        td = trader.get_trading_days(start_date=start_date, end_date=end_date)
    else:
        td = pd.DataFrame(index=pd.date_range(start_date, end_date))

    td["equity"] = 0.0

    cash_df = await get_cash(portfolio.account_id, portfolio.portfolio_size)

    symbols = trades.symbol.unique().tolist()
    for i in tqdm(
        range(len(symbols)), desc=f"loading portfolio {portfolio_id}"
    ):
        symbol = symbols[i]
        symbol_trades = trades[trades.symbol == symbol].sort_values(
            by="client_time"
        )
        calc_symbol_trades_returns(symbol, symbol_trades, td, data_loader)
    td = td.join(cash_df)
    td = td.fillna(method="ffill")
    td["totals"] = td["equity"] + td["cash"]
    return pd.DataFrame(td, columns=["equity", "cash", "totals"])


async def calc_hyperparameters_analysis(optimizer_run_id: str) -> pd.DataFrame:
    portfolio_ids_parameters = await OptimizerRun.get_portfolio_ids_parameters(
        optimizer_run_id
    )

    portfolio_ids, hypers = list(map(list, zip(*portfolio_ids_parameters)))
    df = None
    if len(portfolio_ids):
        for i in tqdm(range(len(portfolio_ids)), desc="Loading Portfolios"):
            _df = await calc_portfolio_returns(portfolio_ids[i])

            if _df.empty:
                continue

            _df["portfolio_id"] = portfolio_ids[i]
            _df["configurations"] = hypers[i]
            _df.reset_index(inplace=True)
            _df = _df.set_index(["portfolio_id", "date"])
            df = pd.concat([df, _df], axis=0) if df is not None else _df

    return df


async def get_portfolio_equity(portfolio_id: str) -> pd.DataFrame:
    await Portfolio.load_by_portfolio_id(portfolio_id)
    data_loader = DataLoader()
    trades = await load_trades_by_portfolio(portfolio_id)

    symbols = trades.symbol.unique().tolist()
    rows = []
    for i in tqdm(
        range(len(symbols)), desc=f"loading portfolio {portfolio_id}"
    ):
        symbol = symbols[i]
        symbol_trades = trades[trades.symbol == symbol].sort_values(
            by="client_time"
        )
        try:
            qty, last_price = calc_symbol_state(
                symbol, symbol_trades, data_loader
            )
        except Exception as e:
            tlog(str(e))
        else:
            rows.append(
                {
                    "symbol": symbol,
                    "qty": qty,
                    "price": last_price,
                    "total": float(qty) * last_price,
                }
            )

    df = pd.DataFrame(rows)
    return df.loc[df.qty > 0].round(2)


async def get_portfolio_cash(portfolio_id: str) -> pd.DataFrame:
    portfolio = await Portfolio.load_by_portfolio_id(portfolio_id)
    return (await Accounts.get_transactions(portfolio.account_id)).round(2)



from typing import List

from liualgotrader.common.types import QueueMapper, WSEventType


class StreamingAPI:
    __instance: object = None

    def __init__(self, queues: QueueMapper):
        self.queues = queues
        StreamingAPI.__instance = self

    async def subscribe(
        self, symbols: List[str], events: List[WSEventType]
    ) -> bool:
        pass

    @classmethod
    def get_instance(cls):
        if not cls.__instance:
            raise AssertionError("Must instantiate before usage")

        return cls.__instance  # type: ignore

    async def unsubscribe(self, symbol: str) -> bool:
        pass

    async def run(self):
        pass

    async def close(self) -> None:
        pass



import asyncio
import base64
import concurrent.futures
import hashlib
import hmac
import json
import os
import queue
import ssl
import time
import traceback
from datetime import date, datetime, timedelta, timezone
from threading import Thread
from typing import Dict, List, Optional

import numpy as np
import pandas as pd
import pytz
import requests
import websocket

from liualgotrader.common import config
from liualgotrader.common.list_utils import chunks
from liualgotrader.common.tlog import tlog
from liualgotrader.common.types import (QueueMapper, TimeScale, Trade,
                                        WSEventType)
from liualgotrader.data.data_base import DataAPI
from liualgotrader.data.streaming_base import StreamingAPI

utctz = pytz.timezone("UTC")


def requests_get(url):
    r = requests.get(url)

    if r.status_code in (429, 502):
        tlog(f"{url} return {r.status_code}, waiting and re-trying")
        time.sleep(10)
        return requests_get(url)

    return r


class GeminiData(DataAPI):
    gemini_api_key: Optional[str] = os.getenv("GEMINI_API_KEY")
    gemini_api_secret: Optional[str] = os.getenv("GEMINI_API_SECRET")
    base_url = "https://api.gemini.com"
    base_websocket = "wss://api.gemini.com"
    datapoints_per_request = 500
    max_trades_per_minute = 10

    def __init__(self):
        self.running_task: Optional[Thread] = None
        self.ws = None

    def get_symbols(self) -> List[str]:
        endpoint = "/v1/symbols"
        url = self.base_url + endpoint
        response = requests.get(url)
        if response.status_code == 200:
            return response.json()

        raise AssertionError(
            f"HTTP ERROR {response.status_code} {response.text}"
        )

    async def get_market_snapshot(self, filter_func) -> List[Dict]:
        raise NotImplementedError

    def _get_ranges(self, start, end):
        start_t = datetime.combine(start, datetime.min.time(), tzinfo=utctz)
        end_t = datetime.combine(end, datetime.max.time(), tzinfo=utctz)

        minutes = (end_t - start_t).total_seconds() / 60
        return pd.date_range(
            start_t,
            end_t,
            periods=minutes
            / (self.datapoints_per_request / self.max_trades_per_minute),
        )

    async def _consolidate_response(self, response, scale) -> pd.DataFrame:
        _df = pd.DataFrame(response.json())

        if _df.empty:
            return pd.DataFrame()

        _df = _df.set_index(_df.timestamp).sort_index()

        _df["s"] = pd.to_datetime(
            (_df.index * 1e9).astype("int64"),
            utc=True,
        )

        # _df.timestamp.apply(lambda x: pd.Timestamp(x, tz=utctz, unit="ns"))
        _df["amount"] = pd.to_numeric(_df.amount)
        _df["price"] = pd.to_numeric(_df.price)
        _df = _df[["s", "price", "amount"]].set_index("s")

        rule = "T" if scale == TimeScale.minute else "D"
        _newdf = _df.resample(rule).first()
        _newdf["high"] = _df.resample(rule).max().price
        _newdf["low"] = _df.resample(rule).min().price
        _newdf["close"] = _df.resample(rule).last().price
        _newdf["count"] = _df.resample(rule).count().amount
        _newdf["volume"] = _df.resample(rule).sum().amount
        _newdf = (
            _newdf.rename(columns={"price": "open", "s": "timestamp"})
            .drop(columns=["amount"])
            .sort_index()
        )
        _newdf = _newdf.dropna()
        _newdf["average"] = 0.0
        _newdf["vwap"] = 0.0

        return _newdf

    async def aget_symbol_data(
        self,
        symbol: str,
        start: date,
        end: date = date.today(),
        scale: TimeScale = TimeScale.minute,
    ) -> pd.DataFrame:
        symbol = symbol.lower()
        tlog(
            f"GEMINI start loading {symbol} from {start} to {end} w scale {scale}"
        )
        ranges = self._get_ranges(start, end)
        endpoint = f"/v1/trades/{symbol}"
        returned_df: pd.DataFrame = pd.DataFrame()
        with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
            loop = asyncio.get_event_loop()
            futures = [
                loop.run_in_executor(
                    executor,
                    requests_get,
                    f"{self.base_url}{endpoint}?timestamp={int(current_timestamp.timestamp())}&limit_trades=500",
                )
                for current_timestamp in ranges[:-1]
            ]

            for response in await asyncio.gather(*futures):
                if response.status_code != 200:
                    raise ValueError(
                        f"HTTP ERROR {response.status_code} {response.text}"
                    )

                df = await self._consolidate_response(response, scale)

                if df.empty:
                    continue
                if returned_df.empty:
                    returned_df = df
                else:
                    returned_df = pd.concat([returned_df, df])
                    returned_df = returned_df[
                        ~returned_df.index.duplicated(keep="first")
                    ]

        tlog(
            f"GEMINI completed loading {symbol} from {start} to {end} w scale {scale}"
        )
        return returned_df

    def get_symbols_data(
        self,
        symbols: List[str],
        start: date,
        end: date = date.today(),
        scale: TimeScale = TimeScale.minute,
    ) -> Dict[str, pd.DataFrame]:
        raise NotImplementedError(
            "get_symbols_data() not implemented yet for Gemini data provider"
        )

    def trading_days_slice(self, symbol: str, s: slice) -> slice:
        return s

    def get_last_trading(self, symbol: str) -> datetime:
        return datetime.now(timezone.utc)

    def get_trading_day(
        self, symbol: str, now: datetime, offset: int
    ) -> datetime:
        return (
            utctz.localize(datetime.combine(now, datetime.min.time()))
            if isinstance(now, date)
            else now
        ) + timedelta(days=offset)

    def get_symbol_data(
        self,
        symbol: str,
        start: date,
        end: date = date.today(),
        scale: TimeScale = TimeScale.minute,
    ) -> pd.DataFrame:
        symbol = symbol.lower()
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(
            self.aget_symbol_data(symbol, start, end, scale)
        )

    def num_trading_minutes(self, symbol: str, start: date, end: date) -> int:
        raise NotImplementedError("num_trading_minutes")

    def num_trading_days(self, symbol: str, start: date, end: date) -> int:
        raise NotImplementedError("num_trading_days")

    def get_max_data_points_per_load(self) -> int:
        raise NotImplementedError("get_max_data_points_per_load")


class GeminiStream(StreamingAPI):
    gemini_api_key: Optional[str] = os.getenv("GEMINI_API_KEY")
    gemini_api_secret: Optional[str] = os.getenv("GEMINI_API_SECRET")
    base_url = "https://api.gemini.com"
    base_websocket = "wss://api.gemini.com"

    def __init__(self, queues: QueueMapper):
        self.running_task: Optional[Thread] = None
        self.ws = None
        super().__init__(queues)

    def _generate_ws_headers(self, payload: Dict) -> Dict:
        if not self.gemini_api_secret or not self.gemini_api_key:
            raise AssertionError(
                "both env variables GEMINI_API_KEY and GEMINI_API_SECRET must be set up"
            )
        t = datetime.now()
        payload_nonce = str(int(time.mktime(t.timetuple()) * 1000))
        payload["nonce"] = payload_nonce
        encoded_payload = json.dumps(payload).encode()
        b64 = base64.b64encode(encoded_payload)
        signature = hmac.new(
            self.gemini_api_secret.encode(), b64, hashlib.sha384
        ).hexdigest()

        return {
            "X-GEMINI-APIKEY": self.gemini_api_key,
            "X-GEMINI-PAYLOAD": b64.decode(),
            "X-GEMINI-SIGNATURE": signature,
        }

    async def run(self):
        if not self.running_task:
            endpoint = "/v1/marketdata/btcusd"  # TODO need support all symbols in an efficient way
            payload = {"request": endpoint}
            headers = self._generate_ws_headers(payload)
            self.ws = websocket.WebSocketApp(
                f"{self.base_websocket}{endpoint}?trades=true&heartbeat=true",
                on_message=self.on_message,
                on_error=self.on_error,
                on_close=self.on_close,
                header=headers,
            )
            self.running_task = Thread(
                target=self.ws.run_forever,
                args=(None, {"cert_reqs": ssl.CERT_NONE}),
            )
            self.running_task.start()

        return self.running_task

    async def close(self):
        if self.running_task and self.running_task.is_alive():
            tlog(f"close task {self.running_task}")
            self.ws.keep_running = False
            self.running_task.join()
            tlog("task terminated")
            self.ws = None
            self.running_task = None

    @classmethod
    def on_message(cls, ws, msgs):
        msg = json.loads(msgs)

        if msg["type"] != "update":
            return

        for event in msg["events"]:
            if event["type"] == "trade":
                cls.trades_handler(
                    pd.Timestamp(
                        datetime.fromtimestamp(msg["timestamp"]).astimezone(
                            utctz
                        )
                    ),
                    event,
                )

    @classmethod
    def on_error(cls, ws, error):
        tlog(f"[ERROR] GeminiStream {error}")

    @classmethod
    def on_close(cls, ws, close_status_code, close_msg):
        tlog(
            f"on_close(): GeminiStream status={close_status_code}, close_msg={close_msg}"
        )

    @classmethod
    def trades_handler(cls, timestamp: datetime, event: Dict):
        try:
            if (time_diff := (datetime.now(timezone.utc) - timestamp)) > timedelta(seconds=2):  # type: ignore
                # if randint(1, 100) == 1:  # nosec
                tlog(f"received a trade for btcusd out of sync w {time_diff}")

            event = {
                "symbol": "btcusd",
                "price": float(event["price"]),
                "open": float(event["price"]),
                "close": float(event["price"]),
                "high": float(event["price"]),
                "low": float(event["price"]),
                "timestamp": timestamp,
                "volume": float(event["amount"]),
                "exchange": "gemini",
                "conditions": event["makerSide"],
                "tape": "",
                "average": None,
                "count": 1,
                "vwap": None,
                "EV": "T",
            }

            cls.get_instance().queues["btcusd"].put(event, block=False)

        except queue.Full as f:
            tlog(
                f"[EXCEPTION] process_message(): queue for {event['sym']} is FULL:{f}"
            )
            raise
        except AssertionError as e:
            tlog(f"[EXCEPTION] GEMINI process_message(): {e}")
            time.sleep(1)
            return
        except Exception as e:
            tlog(
                f"[EXCEPTION] process_message(): exception of type {type(e).__name__} with args {e.args}"
            )
            if config.debug_enabled:
                traceback.print_exc()

    async def subscribe(
        self, symbols: List[str], events: List[WSEventType]
    ) -> bool:
        return True  # TODO Gemini - handle all symbols and event types



from liualgotrader.common import config
from liualgotrader.common.tlog import tlog


class KeyStore:
    @classmethod
    async def load(cls, key: str):
        pool = config.db_conn_pool
        async with pool.acquire() as con:
            val = await con.fetchval(
                """
                    SELECT 
                        value
                    FROM 
                        keystore
                    WHERE 
                        key = $1
                """,
                key,
            )

            return val

    @classmethod
    async def save(cls, key: str, value: str):
        pool = config.db_conn_pool
        async with pool.acquire() as con:
            await con.execute(
                """
                    INSERT INTO 
                        keystore(key, value)
                    VALUES 
                        ($1, $2)
                    ON CONFLICT(key) 
                        DO UPDATE 
                            SET value = $2;
                """,
                key,
                value,
            )



from dataclasses import dataclass
from enum import Enum, auto
from typing import Dict, List, Optional

import pandas as pd
from mnqueues import MNQueue


class DataConnectorType(Enum):
    polygon = 1
    alpaca = 2
    finnhub = 3
    gemini = 4
    tradier = 5


class BrokerType(Enum):
    alpaca = 1
    gemini = 2
    tradier = 3


class TimeScale(Enum):
    day = 24 * 60 * 60
    minute = 60


class WSEventType(Enum):
    TRADE = auto()
    QUOTE = auto()
    MIN_AGG = auto()
    SEC_AGG = auto()


class WSConnectState(Enum):
    NOT_CONNECTED = auto()
    CONNECTED = auto()
    AUTHENTICATED = auto()


class AssetType(Enum):
    US_EQUITIES = auto()
    CRYPTO = auto()


@dataclass
class Order:
    class EventType(Enum):
        partial_fill = auto()
        fill = auto()
        canceled = auto()
        rejected = auto()
        cancel_rejected = auto()
        pending = auto()
        error = auto()
        open = auto()
        other = auto()

    class FillSide(Enum):
        buy = auto()
        sell = auto()

    order_id: str
    symbol: str
    event: EventType
    submitted_at: pd.Timestamp
    price: Optional[float] = None
    trade_fees: Optional[float] = None
    filled_qty: Optional[float] = None
    side: Optional[FillSide] = None
    remaining_amount: Optional[float] = None
    avg_execution_price: Optional[float] = None
    external_account_id: Optional[str] = None


@dataclass
class ThreadFlags:
    run: bool = True


@dataclass
class Trade:
    order_id: str
    symbol: str
    event: Order.EventType
    side: Order.FillSide
    filled_qty: float
    trade_fee: float
    filled_avg_price: float
    liquidity: str
    updated_at: pd.Timestamp


class QueueMapper:
    def __init__(self, queue_list: List[MNQueue] = None):
        self.queues: Dict[str, MNQueue] = {}
        self.queue_list: Optional[List[MNQueue]] = queue_list

    def __repr__(self):
        return str(list(self.queues.keys()))

    def __getitem__(self, key: str) -> MNQueue:
        try:
            return self.queues[key.lower()]
        except KeyError:
            raise AssertionError(f"No queue exists for symbol {key}")

    def __setitem__(self, key: str, newvalue: MNQueue):
        if self.queue_list and newvalue not in self.queue_list:
            raise AssertionError(f"key {key} added to unknown Queue")
        self.queues[key.lower()] = newvalue

    def get_allqueues(self) -> Optional[List[MNQueue]]:
        return self.queue_list



import asyncio
from datetime import date, datetime, timedelta
from typing import Callable, List, Optional, Tuple

import requests
from alpaca_trade_api.rest import REST as tradeapi
from pytz import timezone

from liualgotrader.common import config
from liualgotrader.common.data_loader import DataLoader  # type: ignore
from liualgotrader.common.tlog import tlog
from liualgotrader.common.types import DataConnectorType
from liualgotrader.data.alpaca import AlpacaData
from liualgotrader.data.polygon import PolygonData
from liualgotrader.models.ticker_data import StockOhlc
from liualgotrader.trading.base import Trader

from .base import Scanner


class Momentum(Scanner):
    name = "momentum"

    def __init__(
        self,
        recurrence: Optional[timedelta],
        target_strategy_name: Optional[str],
        data_loader: DataLoader,
        trading_api: Trader,
        max_share_price: float,
        min_share_price: float,
        min_last_dv: float,
        today_change_percent: float,
        min_volume: float,
        from_market_open: float,
        max_symbols: int = config.total_tickers,
        data_source: object = None,
    ):
        self.max_share_price = max_share_price
        self.min_share_price = min_share_price
        self.min_last_dv = min_last_dv
        self.min_volume = min_volume
        self.today_change_percent = today_change_percent
        self.from_market_open = from_market_open
        self.max_symbols = max_symbols
        self.trading_api = trading_api

        super().__init__(
            name=self.name,
            recurrence=recurrence,
            target_strategy_name=target_strategy_name,
            data_loader=data_loader,
            data_source=data_source,
        )

    @classmethod
    def __str__(cls) -> str:
        return cls.name

    async def _get_trade_able_symbols(self) -> List[str]:
        symbols = await self.trading_api.get_tradeable_symbols()
        tlog(
            f"loaded list of {len(symbols)} trade-able symbols from {self.trading_api}"
        )
        return symbols

    async def apply_filter_on_market_snapshot(
        self, filter_func: Optional[Callable]
    ) -> List[str]:
        filtered = await self.data_loader.data_api.get_market_snapshot(
            filter_func
        )
        tlog(
            f"loaded {len(filtered)} tickers of market snapshots after momentum filtering"
        )
        if not len(filtered):
            tlog("failed to load any market snapshots for any tickers")
            return []

        return [x["ticker"] for x in filtered]

    async def add_stock_data_for_date(self, symbol: str, when: date) -> None:
        _minute_data = self.data_loader[symbol][
            when : when + timedelta(days=1)  # type: ignore
        ]

        await asyncio.sleep(0)
        if _minute_data[symbol].empty:
            daily_bar = StockOhlc(
                symbol=symbol,
                symbol_date=when,
                open=0.0,
                high=0.0,
                low=0.0,
                close=0.0,
                volume=0,
                indicators={},
            )
            await daily_bar.save()
        else:
            for index, row in _minute_data[symbol].iterrows():
                daily_bar = StockOhlc(
                    symbol=symbol,
                    symbol_date=index,
                    open=row["open"],
                    high=row["high"],
                    low=row["low"],
                    close=row["close"],
                    volume=int(row["volume"]),
                    indicators={},
                )
                await daily_bar.save()
                print(f"saved data for {symbol} @ {index}")

    async def load_from_db(self, back_time: datetime) -> List[str]:
        pool = config.db_conn_pool

        daily_scale = back_time.hour == 0
        start = (
            back_time
            if daily_scale
            else back_time.replace(hour=9, minute=30, second=0, microsecond=0)
        )
        end = (
            back_time + timedelta(days=1)
            if daily_scale
            else back_time + timedelta(minutes=1)
        )
        async with pool.acquire() as con:
            rows = await con.fetch(
                """
                    SELECT
                        symbol
                    FROM 
                        trending_tickers
                    WHERE
                        scanner_name = $1 
                        AND create_tstamp >= $2
                        AND create_tstamp <= $3
                """,
                self.name,
                start,
                end,
            )

            print("load from db", start, end, len(rows))
            return [row[0] for row in rows] if len(rows) > 0 else []

    def momentum_filter(self, snapshot) -> bool:
        if isinstance(self.data_loader.data_api, AlpacaData):
            (
                tradable,
                last_trade_price,
                prev_day_volume_price,
                today_gap_up_pct,
                today_volume,
            ) = self.alpaca_snapshot_details(snapshot)
        elif isinstance(self.data_loader.data_api, PolygonData):
            (
                tradable,
                last_trade_price,
                prev_day_volume_price,
                today_gap_up_pct,
                today_volume,
            ) = self.polygon_snapshot_details(snapshot)
        else:
            raise ValueError(
                f"Invalid data API: {type(self.data_loader.data_api)}"
            )

        in_price_range = (
            self.max_share_price >= last_trade_price >= self.min_share_price
        )
        return (
            tradable
            and in_price_range
            and prev_day_volume_price > self.min_last_dv
            and today_gap_up_pct >= self.today_change_percent
            and today_volume > self.min_volume
        )

    def polygon_snapshot_details(
        self, snapshot
    ) -> Tuple[bool, float, float, float, float]:
        tradeable = snapshot["ticker"].lower() in self.trade_able_symbols
        last_trade_price = float(snapshot["lastTrade"]["p"])
        prev_day_volume_price = float(snapshot["prevDay"]["p"]) * float(
            snapshot["prevDay"]["v"]
        )
        today_gap_up_pct = snapshot["todaysChangePerc"]
        today_volume = snapshot["day"]["v"]

        return (
            tradeable,
            last_trade_price,
            prev_day_volume_price,
            today_gap_up_pct,
            today_volume,
        )

    def alpaca_snapshot_details(
        self, snapshot
    ) -> Tuple[bool, float, float, float, float]:
        tradable = snapshot["ticker"].lower() in self.trade_able_symbols

        last_trade_price = float(snapshot["latest_trade"]["p"])
        prev_day_price_volume = float(snapshot["prev_daily_bar"]["v"]) * float(
            snapshot["prev_daily_bar"]["c"]
        )
        today_gap_up_pct = (
            100.0
            * (snapshot["daily_bar"]["o"] - snapshot["prev_daily_bar"]["c"])
            / snapshot["prev_daily_bar"]["c"]
        )
        today_volume = snapshot["daily_bar"]["v"]

        return (
            tradable,
            last_trade_price,
            prev_day_price_volume,
            today_gap_up_pct,
            today_volume,
        )

    async def run(self, back_time: datetime = None) -> List[str]:
        if not back_time:
            self.trade_able_symbols = await self._get_trade_able_symbols()
            return await self.apply_filter_on_market_snapshot(
                self.momentum_filter
            )

        rows = await self.load_from_db(back_time)

        tlog(
            f"Scanner {self.name} -> back_time={back_time} picked {len(rows)}"
        )
        return rows



"""
Run scanners periodically, and pump newly scanned symbols to the Producer process
"""
import asyncio
import json
import multiprocessing as mp
import os
import traceback
from datetime import timedelta
from typing import Dict, List

from pytz import timezone

from liualgotrader.common import config
from liualgotrader.common.data_loader import DataLoader  # type: ignore
from liualgotrader.common.database import create_db_connection
from liualgotrader.common.tlog import tlog
from liualgotrader.common.types import DataConnectorType
from liualgotrader.scanners.base import Scanner
from liualgotrader.scanners.momentum import Momentum
from liualgotrader.trading.base import Trader
from liualgotrader.trading.trader_factory import trader_factory

nyc = timezone("America/New_York")


async def scanner_runner(scanner: Scanner, queue: mp.Queue) -> None:
    try:
        while True:
            symbols = await scanner.run()

            if len(symbols):
                tlog(f"Scanner {scanner.name} picked {len(symbols)} symbols")
                queue.put(
                    json.dumps(
                        [
                            {
                                "symbol": symbol.lower(),
                                "target_strategy_name": scanner.target_strategy_name,
                            }
                            for symbol in symbols
                        ]
                    )
                )

            if not scanner.recurrence:
                break

            await asyncio.sleep(scanner.recurrence.total_seconds())
            tlog(f"scanner {scanner.name} re-running")
    except asyncio.CancelledError:
        tlog(
            f"scanner_runner() cancelled, closing scanner task {scanner.name}"
        )
    except Exception as e:
        traceback.print_exc()
        tlog(
            f"[ERROR]Exception in scanner_runner({scanner.name}): exception of type {type(e).__name__} with args {e.args}"
        )
    finally:
        tlog(f"scanner_runner {scanner.name} completed")


async def create_momentum_scanner(
    trader: Trader, data_loader: DataLoader, scanner_details: Dict
) -> Momentum:
    try:
        recurrence = scanner_details.get("recurrence", None)
        target_strategy_name = scanner_details.get(
            "target_strategy_name", None
        )
        return Momentum(
            data_loader=data_loader,
            trading_api=trader,
            min_last_dv=scanner_details["min_last_dv"],
            min_share_price=scanner_details["min_share_price"],
            max_share_price=scanner_details["max_share_price"],
            min_volume=scanner_details["min_volume"],
            from_market_open=scanner_details["from_market_open"],
            today_change_percent=scanner_details["today_change_percent"],
            recurrence=timedelta(minutes=recurrence) if recurrence else None,
            target_strategy_name=target_strategy_name,
            max_symbols=scanner_details.get(
                "max_symbols", config.total_tickers
            ),
        )
    except KeyError as e:
        tlog(
            f"Error {e} in processing of scanner configuration {scanner_details}"
        )
        exit(0)


async def create_scanners(
    trader: Trader, data_loader: DataLoader, scanners_conf: Dict
) -> List[Scanner]:
    scanners: List[Scanner] = []

    for scanner_name in scanners_conf:
        if scanner_name == "momentum":
            scanners.append(
                await create_momentum_scanner(
                    trader,
                    DataLoader(connector=data_loader),
                    scanners_conf[scanner_name],
                )
            )
            tlog("instantiated momentum scanner")
        else:
            tlog(f"custom scanner {scanner_name} selected")
            scanners.append(
                await Scanner.get_scanner(
                    data_loader, scanner_name, scanners_conf[scanner_name]
                )
            )

    return scanners


async def scanners_runner(
    scanners_conf: Dict, queue: mp.Queue, trader: Trader
) -> None:
    print("** scanners_runner() task starting **")
    scanners: List[Scanner] = await create_scanners(
        trader, DataLoader(), scanners_conf
    )
    scanner_tasks = [
        asyncio.create_task(scanner_runner(scanner, queue))
        for scanner in scanners
    ]

    try:
        await asyncio.gather(
            *scanner_tasks,
            return_exceptions=True,
        )

    except asyncio.CancelledError:
        tlog(
            "scanners_runner.scanners_runner() cancelled, closing scanner tasks"
        )

        for task in scanner_tasks:
            tlog(
                f"scanners_runner.scanners_runner()  requesting task {task.get_name()} to cancel"
            )
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                tlog(
                    "scanners_runner.scanners_runner()  task is cancelled now"
                )

    finally:
        queue.close()
        tlog("scanners_runner.scanners_runner()  done.")


async def async_main(scanners_conf: Dict, queue: mp.Queue) -> None:
    await create_db_connection(str(config.dsn))

    main_task = asyncio.create_task(
        scanners_runner(
            scanners_conf,
            queue,
            trader=(at := trader_factory()),
        ),
        name="main_task",
    )

    await asyncio.gather(
        main_task,
        return_exceptions=True,
    )


def main(
    conf_dict: Dict,
    scanner_queue: mp.Queue,
) -> None:
    tlog(f"*** scanners_runner.main() starting w pid {os.getpid()} ***")

    if scanners_conf := conf_dict["scanners"]:
        try:
            asyncio.run(async_main(scanners_conf, scanner_queue))
        except KeyboardInterrupt:
            tlog("scanners_runner.main() - Caught KeyboardInterrupt")
        except Exception as e:
            tlog(
                f"scanners_runner.main() - exception of type {type(e).__name__} with args {e.args}"
            )

    tlog("*** scanners_runner.main() completed ***")



import json
from datetime import datetime
from typing import List, Optional

import asyncpg

from liualgotrader.common import config


class TradePlan:
    def __init__(
        self,
        strategy_name: str,
        parameters: dict,
        portfolio_id: str,
    ):
        self.strategy_name = strategy_name
        self.parameters = parameters
        self.portfolio_id = portfolio_id

    def __str__(self):
        return f"{self.strategy_name}:{self.parameters}"

    @classmethod
    async def load(cls) -> List["TradePlan"]:
        pool = config.db_conn_pool
        async with pool.acquire() as con:
            async with con.transaction():
                rows = await con.fetch(
                    """
                        SELECT
                            strategy_name, portfolio_id, parameters
                        FROM 
                            trade_plan
                        WHERE 
                            expire_tstamp is null
                            AND start_date <= NOW()
                    """,
                )

                return [
                    TradePlan(
                        strategy_name=r["strategy_name"],
                        portfolio_id=r["portfolio_id"],
                        parameters=json.loads(r["parameters"]),
                    )
                    for r in rows
                ]

    @classmethod
    async def get_new_entries(
        cls, since: datetime
    ) -> Optional[List["TradePlan"]]:
        pool = config.db_conn_pool
        async with pool.acquire() as con:
            q = f"""
                    SELECT
                        strategy_name, portfolio_id, parameters
                    FROM 
                        trade_plan
                    WHERE 
                        expire_tstamp is null
                        AND start_date <= NOW()
                        AND create_tstamp > '{since}'
                """
            rows = await con.fetch(q)

            return [
                TradePlan(
                    strategy_name=r["strategy_name"],
                    portfolio_id=r["portfolio_id"],
                    parameters=json.loads(r["parameters"]),
                )
                for r in rows
            ]



from datetime import date

from pandas import DataFrame

from liualgotrader.common import config
from liualgotrader.common.database import fetch_as_dataframe
from liualgotrader.common.tlog import tlog


class GainLoss:
    @classmethod
    async def save(cls, df: DataFrame):
        pool = config.db_conn_pool

        async with pool.acquire() as con:
            for _, row in df.iterrows():
                try:
                    async with con.transaction():
                        _ = await con.fetchval(
                            """
                                INSERT INTO gain_loss (symbol, algo_run_id, gain_percentage, gain_value)
                                VALUES ($1, $2, $3, $4)
                                RETURNING gain_loss_id
                            """,
                            row.symbol,
                            row.algo_run_id,
                            row.gain_percentage,
                            row.gain_value,
                        )
                except Exception as e:
                    tlog(f"[ERROR] inserting {row} resulted in exception {e}")

    @classmethod
    async def load(cls, start_date: date) -> DataFrame:
        q = """
            SELECT symbol, algo_name, algo_env, start_time, gain_percentage, gain_value
            FROM gain_loss as g, algo_run as a
            WHERE 
                g.algo_run_id = a.algo_run_id AND
                start_time >= $1
            ORDER BY symbol, algo_name, start_time
            """

        return await fetch_as_dataframe(q, start_date)


class TradeAnalysis:
    @classmethod
    async def save(cls, df: DataFrame):
        pool = config.db_conn_pool

        async with pool.acquire() as con:
            for _, row in df.iterrows():
                try:
                    async with con.transaction():
                        _ = await con.fetchval(
                            """
                                INSERT INTO trade_analysis (symbol, algo_run_id, gain_percentage, gain_value, r_units, start_tstamp, end_tstamp)
                                VALUES ($1, $2, $3, $4, $5, $6, $7)
                                RETURNING trade_analysis_id
                            """,
                            row.symbol,
                            row.algo_run_id,
                            row.gain_percentage,
                            row.gain_value,
                            row.r_units,
                            row.start_time.to_pydatetime(),
                            row.end_time.to_pydatetime(),
                        )
                except Exception as e:
                    tlog(f"[ERROR] inserting {row} resulted in exception {e}")

    @classmethod
    async def load(cls, start_date: date) -> DataFrame:
        q = """
            SELECT symbol, algo_name, algo_env, r_units, gain_percentage, gain_value, t.start_tstamp, t.end_tstamp
            FROM trade_analysis as t, algo_run as a
            WHERE 
                t.algo_run_id = a.algo_run_id AND
                start_time >= $1
            ORDER BY symbol, algo_name, start_time
            """

        return await fetch_as_dataframe(q, start_date)



"""
Execute Strategies on streaming data received from the Producer
"""
import asyncio
import os
from datetime import datetime, timedelta
from queue import Empty
from random import randint
from typing import Any, Dict, List, Optional

import pandas as pd
import pygit2
from mnqueues import MNQueue
from pandas import DataFrame as df
from pytz import timezone

from liualgotrader.analytics.analysis import load_trades_by_portfolio
from liualgotrader.common import config, market_data, trading_data
from liualgotrader.common.data_loader import DataLoader  # type: ignore
from liualgotrader.common.database import create_db_connection
from liualgotrader.common.tlog import tlog, tlog_exception
from liualgotrader.common.types import Order, Trade
from liualgotrader.fincalcs.data_conditions import QUOTE_SKIP_CONDITIONS
from liualgotrader.models.new_trades import NewTrade
from liualgotrader.models.portfolio import Portfolio
from liualgotrader.models.tradeplan import TradePlan
from liualgotrader.models.trending_tickers import TrendingTickers
from liualgotrader.strategies.base import Strategy, StrategyType
from liualgotrader.trading.base import Trader
from liualgotrader.trading.trader_factory import (get_trader_by_name,
                                                  trader_factory)

shortable: Dict = {}
symbol_data_error: Dict = {}
rejects: Dict[str, List[str]] = {}
time_tick: Dict[str, datetime] = {}
nyc = timezone("America/New_York")


async def end_time(reason: str):
    for s in trading_data.strategies:
        tlog(f"updating end time for strategy {s.name}")
        await s.algo_run.update_end_time(
            pool=config.db_conn_pool, end_reason=reason
        )


def get_position(trader: Trader, symbol: str) -> float:
    try:
        return trader.get_position(symbol)
    except Exception:
        return 0


async def execute_run_all_results(
    strategy: Strategy,
    run_all_results: Dict[str, Dict],
    trader: Trader,
    data_loader: DataLoader,
):
    external_account_id = None
    if hasattr(strategy, "portfolio_id"):
        (
            external_account_id,
            broker_name,
        ) = await Portfolio.get_external_account_id(
            strategy.portfolio_id  # type: ignore
        )
        if broker_name:
            trader = get_trader_by_name(broker_name)

    for symbol, what in run_all_results.items():
        await execute_strategy_result(
            strategy,
            trader,
            data_loader,
            symbol.lower(),
            what,
            external_account_id,
        )


async def do_strategy_all(
    data_loader: DataLoader,
    trader: Trader,
    strategy: Strategy,
    symbols: List[str],
    carrier=None,
):
    try:
        now = datetime.now(nyc)
        symbols_position = {
            symbol.lower(): trading_data.positions[symbol]
            if symbol in trading_data.positions
            else get_position(trader, symbol)
            for symbol in symbols
        }
        do = await strategy.run_all(
            symbols_position=symbols_position,
            now=now,
            portfolio_value=config.portfolio_value,
            backtesting=True,
            data_loader=data_loader,
            trader=trader,
        )
        await execute_run_all_results(strategy, do, trader, data_loader)

    except Exception as e:
        if config.debug_enabled:
            tlog_exception("do_strategy_all")
        tlog(f"[Exception] {now} {strategy}->{e}")

        raise


async def cancel_lingering_orders(trader: Trader):
    tlog("cancel_lingering_orders() task starting")

    while True:
        await asyncio.sleep(60)

        if not len(trading_data.open_orders):
            continue

        ny_now = datetime.now(nyc)
        if not trader.is_market_open(ny_now):
            tlog("cancel_lingering_orders() market is closed")
            continue

        t = list(trading_data.open_orders.items())
        for symbol, order in t:
            await order_inflight(
                symbol=symbol,
                existing_order=order,
                now=ny_now,
                trader=trader,
            )


async def periodic_runner(data_loader: DataLoader, trader: Trader) -> None:
    try:
        while True:
            tlog("periodic_runner() task starting")
            # run strategies
            tasks = []
            for s in trading_data.strategies:
                try:
                    skip = not await s.should_run_all()
                except Exception:
                    skip = True
                finally:
                    if skip:
                        continue

                symbols = [
                    symbol.lower()
                    for symbol in trading_data.last_used_strategy
                    if trading_data.last_used_strategy[symbol.lower()] == s
                ]

                tasks.append(
                    asyncio.create_task(
                        do_strategy_all(
                            trader=trader,
                            data_loader=data_loader,
                            strategy=s,
                            symbols=symbols,
                        )
                    )
                )

            asyncio.gather(*tasks)
            await asyncio.sleep(5 * 60.0)

    except asyncio.CancelledError:
        tlog("periodic_runner() cancelled")
    except KeyboardInterrupt:
        tlog("periodic_runner() - Caught KeyboardInterrupt")
    except Exception as e:
        if config.debug_enabled:
            tlog_exception("periodic_runner")
        tlog(f"[EXCEPTION] periodic_runner: {e}")

    tlog("periodic_runner() task completed")


async def should_cancel_order(order: Order, market_clock: datetime) -> bool:
    # Make sure the order's not too old
    submitted_at = order.submitted_at.astimezone(market_clock.tzinfo)
    order_lifetime = market_clock - submitted_at

    if config.debug_enabled:
        tlog(
            f"should_cancel_order submitted_at:{submitted_at}, order_lifetime:{order_lifetime}"
        )
    return (
        market_clock > submitted_at
        and order_lifetime.total_seconds() // 60 >= 1
    )


async def save(
    symbol: str,
    new_qty: float,
    last_op: str,
    price: float,
    indicators: Dict[Any, Any],
    now: str,
    trade_fee=0.0,
) -> None:
    symbol = symbol.lower()
    db_trade = NewTrade(
        algo_run_id=trading_data.open_order_strategy[symbol].algo_run.run_id,
        symbol=symbol,
        qty=new_qty,
        operation=last_op,
        price=price,
        indicators=indicators,
    )

    await db_trade.save(
        config.db_conn_pool,
        now,
        trading_data.stop_prices[symbol]
        if symbol in trading_data.stop_prices
        else 0.0,
        trading_data.target_prices[symbol]
        if symbol in trading_data.target_prices
        else 0.0,
        trade_fee,
    )


async def do_callbacks(
    symbol: str,
    strategy: Strategy,
    filled_qty: float,
    side: Order.FillSide,
    filled_avg_price: float,
):
    symbol = symbol.lower()
    if side == Order.FillSide.buy:
        trading_data.buy_indicators.pop(symbol, None)
        if strategy:
            await strategy.buy_callback(symbol, filled_avg_price, filled_qty)
    else:
        trading_data.sell_indicators.pop(symbol, None)
        if strategy:
            await strategy.sell_callback(symbol, filled_avg_price, filled_qty)


async def update_partially_filled_order(
    symbol: str,
    strategy: Strategy,
    filled_qty: float,
    side: Order.FillSide,
    filled_avg_price: float,
    updated_at: pd.Timestamp,
    trade_fee: float,
) -> None:
    symbol = symbol.lower()

    await do_callbacks(
        symbol=symbol,
        strategy=strategy,
        filled_qty=filled_qty,
        side=side,
        filled_avg_price=filled_avg_price,
    )

    trading_data.positions[symbol] = round(
        trading_data.positions.get(symbol, 0.0)
        + filled_qty * (1 if side == Order.FillSide.buy else -1),
        8,
    )

    try:
        indicators = {
            "buy": trading_data.buy_indicators.get(symbol, None),
            "sell": trading_data.sell_indicators.get(symbol, None),
        }
    except KeyError:
        indicators = {}

    await save(
        symbol,
        filled_qty,
        side.name,
        filled_avg_price,
        indicators,
        updated_at,
        trade_fee,
    )


async def update_filled_order(
    symbol: str,
    strategy: Strategy,
    filled_qty: float,
    side: Order.FillSide,
    filled_avg_price: float,
    updated_at: pd.Timestamp,
    trade_fee: float,
) -> None:
    symbol = symbol.lower()

    await update_partially_filled_order(
        symbol=symbol,
        strategy=strategy,
        filled_qty=filled_qty,
        side=side,
        filled_avg_price=filled_avg_price,
        updated_at=updated_at,
        trade_fee=trade_fee,
    )
    trading_data.open_orders.pop(symbol)
    if symbol in trading_data.open_order_strategy:
        trading_data.open_order_strategy.pop(symbol)

    tlog(
        f"update_filled_order open order for {symbol} popped. Position now {trading_data.positions[symbol]}"
    )


async def handle_trade_update_for_order(trade: Trade) -> bool:
    symbol = trade.symbol.lower()
    event = trade.event

    tlog(f"trade update for {symbol} with event {event}")

    if event == Order.EventType.partial_fill:
        await update_partially_filled_order(
            symbol=symbol,
            strategy=trading_data.open_order_strategy[symbol],
            filled_qty=trade.filled_qty,
            filled_avg_price=trade.filled_avg_price,
            side=trade.side,
            updated_at=trade.updated_at,
            trade_fee=trade.trade_fee,
        )

    elif event == Order.EventType.fill:
        await update_filled_order(
            symbol=symbol,
            strategy=trading_data.open_order_strategy[symbol],
            filled_qty=trade.filled_qty,
            filled_avg_price=trade.filled_avg_price,
            side=trade.side,
            updated_at=trade.updated_at,
            trade_fee=trade.trade_fee,
        )

    else:
        trading_data.partial_fills.pop(symbol, None)
        trading_data.partial_fills_fee.pop(symbol, None)
        trading_data.open_orders.pop(symbol, None)
        trading_data.open_order_strategy.pop(symbol, None)

    return True


async def handle_trade_update_wo_order(trade: Trade) -> bool:
    trade.symbol.lower()
    trade.event
    # tlog(f"trade update without order for {symbol} data={trade} with event {event}")
    return True


async def handle_trade_update(trade: Trade) -> bool:
    if trade.symbol.lower() in trading_data.open_orders:
        return await handle_trade_update_for_order(trade)
    else:
        return await handle_trade_update_wo_order(trade)


async def handle_quote(data: Dict) -> bool:
    if "askprice" not in data or "bidprice" not in data:
        return True
    if "condition" in data and data["condition"] in QUOTE_SKIP_CONDITIONS:
        return True

    symbol = data["symbol"].lower()
    # tlog(f"quote={data}")
    prev_ask = trading_data.voi_ask.get(symbol, None)
    prev_bid = trading_data.voi_bid.get(symbol, None)
    trading_data.voi_ask[symbol] = (
        data["askprice"],
        data["asksize"],
        data["timestamp"],
    )
    trading_data.voi_bid[symbol] = (
        data["bidprice"],
        data["bidsize"],
        data["timestamp"],
    )

    bid_delta_volume = (
        0
        if not prev_bid or data["bidprice"] < prev_bid[0]
        else 100 * data["bidsize"]
        if data["bidprice"] > prev_bid[0]
        else 100 * (data["bidsize"] - prev_bid[1])
    )
    ask_delta_volume = (
        0
        if not prev_ask or data["askprice"] > prev_ask[0]
        else 100 * data["asksize"]
        if data["askprice"] < prev_ask[0]
        else 100 * (data["asksize"] - prev_ask[1])
    )
    voi_stack = trading_data.voi.get(symbol, None)
    if not voi_stack:
        voi_stack = [0.0]
    elif len(voi_stack) == 10:
        voi_stack[0:9] = voi_stack[1:10]
        voi_stack.pop()

    k = 2.0 / (100 + 1)
    voi_stack.append(
        round(
            voi_stack[-1] * (1.0 - k)
            + k * (bid_delta_volume - ask_delta_volume),
            2,
        )
    )
    trading_data.voi[symbol] = voi_stack
    # tlog(f"{symbol} voi:{trading_data.voi[symbol]}")

    return True


async def aggregate_bar_data(
    data_loader: DataLoader, data: Dict, ts: pd.Timestamp, carrier=None
) -> None:
    symbol = data["symbol"].lower()

    if not data_loader.exist(symbol):
        # print(f"loading data for {symbol}...")
        data_loader[symbol][-1]

    try:
        current = data_loader[symbol].loc[ts]
    except KeyError:
        current = None

    if current is None or data["EV"] == "AM":
        new_data = [
            data["open"],
            data["high"],
            data["low"],
            data["close"],
            data["volume"],
            data["vwap"],
            data["average"],
            data["count"],
        ]
    else:
        new_data = [
            current["open"],
            max(data["high"], current["high"]),
            min(data["low"], current["low"]),
            data["close"],
            current["volume"] + data["volume"],
            data["vwap"] or current["vwap"],
            data["average"] or current["average"],
            data["count"] + current["count"],
        ]
    try:
        data_loader[symbol].loc[ts] = new_data
    except ValueError:
        print(f"loaded for {symbol} {new_data}")
        data_loader[symbol][-1]
        data_loader[symbol].loc[ts] = new_data

    market_data.volume_today[symbol] = (
        market_data.volume_today[symbol] + data["volume"]
        if symbol in market_data.volume_today
        else data["volume"]
    )


async def order_inflight(
    symbol: str,
    existing_order: Order,
    now: pd.Timestamp,
    trader: Trader,
) -> None:
    symbol = symbol.lower()
    try:
        if await should_cancel_order(existing_order, now):
            if config.debug_enabled:
                tlog(
                    f"should_cancel_order - checking status w/ order-id {existing_order.order_id}"
                )
            (
                order_status,
                filled_price,
                filled_qty,
                trade_fee,
            ) = await trader.is_order_completed(
                existing_order.order_id, existing_order.external_account_id
            )

            if order_status == Order.EventType.fill:
                tlog(
                    f"order_id {existing_order.order_id} for {symbol} already filled"
                )
                await update_filled_order(
                    symbol=symbol,
                    strategy=trading_data.open_order_strategy[symbol],  # type: ignore
                    filled_avg_price=filled_price,  # type: ignore
                    filled_qty=filled_qty,  # type: ignore
                    side=existing_order.side,  # type: ignore
                    updated_at=existing_order.submitted_at,  # type: ignore
                    trade_fee=trade_fee,  # type: ignore
                )
            elif order_status == Order.EventType.partial_fill:
                tlog(
                    f"order_id {existing_order.id} for {symbol} already partially_filled"  # type: ignore
                )
                await update_partially_filled_order(
                    symbol=symbol,
                    strategy=trading_data.open_order_strategy[symbol],
                    side=existing_order.side,  # type: ignore
                    filled_avg_price=filled_price,  # type: ignore
                    filled_qty=filled_qty,  # type: ignore
                    updated_at=existing_order.submitted_at,
                    trade_fee=trade_fee,  # type: ignore
                )
            else:
                # Cancel it so we can try again for a fill
                tlog(
                    f"Cancel order id {existing_order.order_id} for {symbol} ts={now} submission_ts={existing_order.submitted_at.astimezone(timezone('America/New_York'))}"  # type: ignore
                )
                if await trader.cancel_order(existing_order):
                    trading_data.open_orders.pop(symbol)

    except AttributeError as e:
        if config.debug_enabled:
            tlog_exception(f"order_inflight() w {e}")
        tlog(f"Attribute Error in symbol {symbol} w/ {existing_order}")
    except Exception as e:
        if config.debug_enabled:
            tlog_exception(f"order_inflight() w {e}")
        tlog(
            f"[EXCEPTION] order_inflight() : {e} for symbol {symbol} w/ {existing_order}"
        )


async def submit_order(
    trader: Trader, symbol: str, what: Dict, external_account_id: str = None
) -> Order:
    return (
        await trader.submit_order(
            symbol=symbol,
            qty=what["qty"],
            side=what["side"],
            order_type="limit",
            time_in_force="day",
            limit_price=what["limit_price"],
            on_behalf_of=external_account_id,
        )
        if what["type"] == "limit"
        else await trader.submit_order(
            symbol=symbol,
            qty=what["qty"],
            side=what["side"],
            order_type=what["type"],
            time_in_force="day",
            on_behalf_of=external_account_id,
        )
    )


async def update_trading_data(
    symbol: str, o: Order, strategy: Strategy, buy: bool
) -> None:
    trading_data.open_orders[symbol] = o
    trading_data.open_order_strategy[symbol] = strategy
    trading_data.last_used_strategy[symbol] = strategy
    if buy:
        trading_data.buy_time[symbol] = datetime.now(
            tz=timezone("America/New_York")
        ).replace(second=0, microsecond=0)


async def execute_strategy_result(
    strategy: Strategy,
    trader: Trader,
    data_loader: DataLoader,
    symbol: str,
    what: Dict,
    external_account_id: str = None,
) -> bool:
    tlog(f"execute_strategy_result for {symbol} do {what}")
    symbol = symbol.lower()

    o = await submit_order(trader, symbol, what, external_account_id)
    if not o:
        return False

    await update_trading_data(symbol, o, strategy, (what["side"] == "buy"))
    await save(
        symbol=symbol,
        new_qty=0,
        last_op=str(what["side"]),
        price=0.0,
        indicators={},
        now=str(datetime.utcnow()),
    )

    if config.debug_enabled:
        tlog(
            f"executed strategy {strategy.name} on {symbol} w data {data_loader[symbol][-10:]}"
        )

    return True


async def do_strategy(
    strategy: Strategy,
    symbol: str,
    shortable: bool,
    position: float,
    data_loader: DataLoader,
    trader: Trader,
    minute_history: df,
    now: pd.Timestamp,
    portfolio_value: Optional[float],
    carrier=None,
) -> bool:
    if not trader.is_market_open(datetime.now(nyc)):
        tlog(f"market closed, can't execute {symbol} @ {strategy.name}")
        return False

    do, what = await strategy.run(
        symbol=symbol,
        shortable=shortable,
        position=position,
        minute_history=minute_history,
        now=now,
        portfolio_value=portfolio_value,
    )

    return (
        await execute_strategy_result(
            strategy=strategy,
            trader=trader,
            data_loader=data_loader,
            symbol=symbol,
            what=what,
        )
        if do
        else not what.get("reject", False)
    )


async def _filter_strategies(symbol: str) -> List:
    return [
        s
        for s in trading_data.strategies
        if not await s.should_run_all()
        and symbol not in rejects.get(s.name, [])
    ]


async def do_strategies(
    trader: Trader,
    data_loader: DataLoader,
    symbol: str,
    position: float,
    now: pd.Timestamp,
    data: Dict,
    carrier=None,
) -> None:
    # run strategies
    strategies = await _filter_strategies(symbol)
    for s in strategies:
        try:
            if not await do_strategy(
                strategy=s,
                symbol=symbol,
                shortable=shortable[symbol],
                position=position,
                minute_history=data_loader[symbol].symbol_data,
                now=pd.to_datetime(now.replace(nanosecond=0)).replace(
                    second=0, microsecond=0
                ),
                portfolio_value=config.portfolio_value,
                data_loader=data_loader,
                trader=trader,
            ):
                if s.name not in rejects:
                    rejects[s.name] = [symbol]
                else:
                    rejects[s.name].append(symbol)

        except Exception as e:
            tlog(f"[EXCEPTION] in do_strategies() : {e}")
            if config.debug_enabled:
                tlog_exception("do_strategies()")


async def handle_aggregate(
    trader: Trader,
    data_loader: DataLoader,
    symbol: str,
    ts: pd.Timestamp,
    data: Dict,
    carrier=None,
) -> bool:
    symbol = symbol.lower()

    if config.debug_enabled:
        tlog(f"handle_aggregate {symbol}")
    # Next, check for existing orders for the stock
    if symbol in trading_data.open_orders and await order_inflight(
        symbol, trading_data.open_orders[symbol.lower()], ts, trader
    ):
        return True

    await do_strategies(
        trader=trader,
        data_loader=data_loader,
        symbol=symbol,
        position=trading_data.positions.get(symbol, 0),
        now=ts,
        data=data,
    )

    return True


async def handle_data_queue_msg(
    data: Dict, trader: Trader, data_loader: DataLoader, carrier=None
) -> bool:
    global shortable
    global symbol_data_error
    global rejects

    ts = pd.to_datetime(
        data["timestamp"].replace(second=0, microsecond=0, nanosecond=0)
    )
    symbol = data["symbol"].lower()
    shortable[symbol] = True  # TODO revisit collecting 'is shortable' data

    # timezone("America/New_York")
    time_diff = datetime.now(tz=data["timestamp"].tz) - data["timestamp"]

    if data["EV"] != "AM" and time_diff > timedelta(seconds=15):
        if randint(1, 100) == 1:  # nosec
            tlog(f"{data['EV']} {symbol} too out of sync w {time_diff}")
        return False

    if (
        time_tick.get(symbol)
        and data["timestamp"].replace(microsecond=0, nanosecond=0)
        == time_tick[symbol]
    ):
        return True

    time_tick[symbol] = data["timestamp"].replace(microsecond=0, nanosecond=0)
    await handle_aggregate(
        trader=trader,
        data_loader=data_loader,
        symbol=symbol,
        ts=time_tick[symbol],
        data=data,
        carrier=None,
    )

    return True


async def queue_consumer(
    batch_id: str, queue: MNQueue, data_loader: DataLoader, trader: Trader
) -> None:
    tlog("queue_consumer() starting")
    try:
        while True:
            try:
                data = queue.get(timeout=2)
                if data["EV"] == "trade_update":
                    tlog(f"received trade_update: {data}")
                    await handle_trade_update(Trade(**data["trade"]))
                elif data["EV"] == "new_strategy":
                    tlog(f"received new_strategy: {data}")
                    await handle_new_strategy(
                        batch_id=batch_id,
                        data_loader=data_loader,
                        portfolio_id=data["portfolio_id"],
                        parameters=data["parameters"],
                    )
                else:
                    await handle_data_queue_msg(data, trader, data_loader)

            except Empty:
                await asyncio.sleep(0)
                continue
            except ConnectionError:
                await trader.reconnect()
                # re-post back to queue
                queue.put(data, timeout=1)
                await asyncio.sleep(1)
                continue
            except Exception as e:
                tlog(
                    f"Exception in queue_consumer(): exception of type {type(e).__name__} with args {e.args} inside loop"
                )
                if config.debug_enabled:
                    tlog_exception("queue_consumer")

    except asyncio.CancelledError:
        tlog("queue_consumer() cancelled ")
    except Exception as e:
        tlog(
            f"Exception in queue_consumer(): exception of type {type(e).__name__} with args {e.args}"
        )

        if config.debug_enabled:
            tlog_exception("queue_consumer")
    finally:
        tlog("queue_consumer() task done.")


async def create_strategies_from_file(
    batch_id: str,
    trader: Trader,
    data_loader: DataLoader,
    strategies_conf: Dict,
) -> List[Strategy]:
    strategy_list = []
    for strategy_name in strategies_conf:
        s = await Strategy.get_strategy(
            batch_id=batch_id,
            strategy_name=strategy_name,
            strategy_details=strategies_conf[strategy_name],
            data_loader=data_loader,
        )
        if s:
            strategy_list.append(s)

    return strategy_list


async def load_symbol_position(portfolio_id: str) -> Dict[str, float]:
    trades = await load_trades_by_portfolio(portfolio_id)

    if not len(trades):
        return {}
    new_df = pd.DataFrame()
    new_df["symbol"] = trades.symbol.unique()
    new_df["qty"] = new_df.symbol.apply(
        lambda x: (
            trades[
                (trades.symbol == x) & (trades.operation == "buy")
            ].qty.sum()
        )
        - trades[(trades.symbol == x) & (trades.operation == "sell")].qty.sum()
    )
    new_df = new_df.loc[new_df.qty != 0]

    return {row.symbol: float(row.qty) for _, row in new_df.iterrows()}


async def create_strategies_from_db(
    batch_id: str,
    trader: Trader,
    data_loader: DataLoader,
) -> List[Strategy]:
    # load from tradeplan file
    trade_plan = await TradePlan.load()

    strategy_list = []
    for trade_plan_entry in trade_plan:
        strategy_details = trade_plan_entry.parameters
        strategy_details["portfolio_id"] = trade_plan_entry.portfolio_id

        strategy_name = strategy_details.pop("name")
        s = await Strategy.get_strategy(
            batch_id=batch_id,
            strategy_name=strategy_name,
            strategy_details=strategy_details,
            data_loader=data_loader,
        )
        if s:
            strategy_list.append(s)
            positions = await load_symbol_position(
                trade_plan_entry.portfolio_id
            )
            tlog(f"Loaded {len(positions)} positions for strategy_name")
            for symbol, qty in positions.items():
                trading_data.positions[symbol] = (
                    trading_data.positions.get(symbol, 0.0) + qty
                )

    return strategy_list


async def handle_new_strategy(
    batch_id: str, portfolio_id: str, parameters: dict, data_loader: DataLoader
):
    strategy_name = parameters.pop("name")
    parameters["portfolio_id"] = portfolio_id
    strategy = await Strategy.get_strategy(
        batch_id=batch_id,
        strategy_name=strategy_name,
        strategy_details=parameters,
        data_loader=data_loader,
    )

    if strategy:
        trading_data.strategies.append(strategy)


async def consumer_async_main(
    queue: MNQueue,
    unique_id: str,
    strategies_conf: Dict,
    file_only: bool,
):
    await create_db_connection(str(config.dsn))
    data_loader = DataLoader()

    trader = trader_factory()

    trading_data.strategies += await create_strategies_from_file(
        batch_id=unique_id,
        trader=trader,
        data_loader=data_loader,
        strategies_conf=strategies_conf,
    )

    if not file_only:
        trading_data.strategies += await create_strategies_from_db(
            batch_id=unique_id,
            trader=trader,
            data_loader=data_loader,
        )

    queue_consumer_task = asyncio.create_task(
        queue_consumer(unique_id, queue, data_loader, trader)
    )

    periodic_runner_task = asyncio.create_task(
        periodic_runner(data_loader, trader)
    )

    cancel_lingering_orders_task = asyncio.create_task(
        cancel_lingering_orders(trader)
    )

    await asyncio.gather(
        queue_consumer_task,
        periodic_runner_task,
        cancel_lingering_orders_task,
        return_exceptions=True,
    )

    tlog("consumer_async_main() completed")


def consumer_main(
    queue: MNQueue,
    unique_id: str,
    conf: Dict,
) -> None:
    tlog(f"*** consumer_main() starting w pid {os.getpid()} ***")

    try:
        config.build_label = pygit2.Repository("../").describe(
            describe_strategy=pygit2.GIT_DESCRIBE_TAGS
        )
    except pygit2.GitError:
        import liualgotrader

        config.build_label = liualgotrader.__version__ if hasattr(liualgotrader, "__version__") else ""  # type: ignore

    config.portfolio_value = conf.get("portfolio_value", None)
    if "risk" in conf:
        config.risk = conf["risk"]

    try:
        asyncio.run(
            consumer_async_main(
                queue,
                unique_id,
                conf["strategies"],
                conf.get("file_only", False),
            ),
        )
    except KeyboardInterrupt:
        tlog("consumer_main() - Caught KeyboardInterrupt")

    tlog("*** consumer_main() completed ***")



"""Base Class for Strategies"""
import importlib
import traceback
from datetime import datetime
from enum import Enum
from typing import Dict, List, Optional, Tuple

from pandas import DataFrame as df

from liualgotrader.common import config
from liualgotrader.common.data_loader import DataLoader  # type: ignore
from liualgotrader.common.tlog import tlog
from liualgotrader.models.accounts import Accounts
from liualgotrader.models.algo_run import AlgoRun
from liualgotrader.models.keystore import KeyStore
from liualgotrader.trading.base import Trader


class StrategyType(Enum):
    DAY_TRADE = 1
    SWING = 2


class Strategy(object):
    def __init__(
        self,
        name: str,
        type: StrategyType,
        batch_id: str,
        schedule: List[Dict],
        ref_run_id: int = None,
        data_loader: DataLoader = None,
        fractional: bool = False,
    ):
        """Strategy default initialization, should be called by all Strategy objects.

        Keyword arguments:
        name: Strategy name, used in the `tradeplan.toml` file,
        type: Strategy type (= day-trade / swing),
        batch_id: generated by the platform, used to group executions,
        schedule: execution time, a defined in the .toml file, note it's mandatory to include,
                  may have more than one execution window. Execution winodows are not enforced
                  by the platform.
        ref_run_id : Used for back-testing,
        data_loader: Passed by the framework, used like a DataFrame to access symbol data.
        fractional: Boolean (default False) if to support fractional trading
        """

        self.name = name
        self.batch_id = batch_id
        self.type = type
        self.ref_run_id = ref_run_id
        self.algo_run = AlgoRun(strategy_name=self.name, batch_id=batch_id)
        self.schedule = schedule
        self.data_loader = data_loader
        self.support_fractional = fractional
        self.account_id = None
        self.global_var: Dict = {}

    def __repr__(self):
        return self.name

    async def calc_qty(
        self, price: float, trade_fee_percentage: float
    ) -> float:
        if not self.account_id:
            raise AssertionError("account_id not set")
        cash = await Accounts.get_balance(self.account_id)
        price *= 1.0 + trade_fee_percentage
        return cash / price if self.support_fractional else cash // price

    async def create(self) -> bool:
        """Called by the framework upon instantiation. Must always call super() implementation.

        Return Values:
        If create returns False, the platform will not execute the Strategy. This is helpful
        specifically for situations when multiple processes are running, yet a strategy
        would like to execute a single copy only.
        """
        await self.algo_run.save(
            pool=config.db_conn_pool, ref_algo_run_id=self.ref_run_id
        )
        return True

    async def should_run_all(self):
        """Called by the framework to select if to use the `run_all()` function, or
        `run()` function.
        """
        return False

    async def run_all(
        self,
        symbols_position: Dict[str, float],
        data_loader: DataLoader,
        now: datetime,
        portfolio_value: float = None,
        trader: Trader = None,
        debug: bool = False,
        backtesting: bool = False,
        fee_buy_percentage: float = 0.0,
        fee_sell_percentage: float = 0.0,
    ) -> Dict[str, Dict]:
        """Called by the framework, periodically, if `should_run_all()` returns True. This function,
        unlike `run()` is executed once, and not per symbol.

        Keyword arguments:
        symbols_position: Dictionary, with open position (quantity) per symbol,
        data_loader: Send by the framework, DataFrame like object to accessing symbol data,
        now: current timestamp (may be in the past, if called by backtester),
        portfolio_value: Sent by the framework [TO BE DEPRECATED],
        trading_api: Sent by the framework, provides access to the Trader [TO BE DEPRECATED],
        debug: Debug flag, used mostly in backtesting,
        backtesting: Flag indicating if calling during backtesting session, or real-time
        fee_buy_percentage: fee to execute buy, presented as % (0-1),
        fee_sell_percentage: fee to execute sell, presented as % (0-1),

        Returns:
        Dictionary with symbol and 'actions' (see documentation for supported actions)
        """
        return {}

    async def run(
        self,
        symbol: str,
        shortable: bool,
        position: float,
        now: datetime,
        minute_history: df,
        portfolio_value: float = None,
        debug: bool = False,
        backtesting: bool = False,
    ) -> Tuple[bool, Dict]:
        """Called by the framework, per symbol.

        Keyword arguments:
        symbol: to act on,
        shortable: [TO BE DEPRECATED],
        position: current position (could be 0),
        now: current time w/ time zone (note can be in the past for backtesting),
        minute_history: DataFrame holding data,
        portfolio_value: [TO BE DEPRECATED],
        debug: Debug flag, used mostly in backtesting,
        backtesting: Flag indicating if calling during backtesting session, or real-time

        Returns:
        False, {} in case no action to be taken,
        True, {action - see documentation for supported actions}
        """
        return False, {}

    async def is_sell_time(self, now: datetime):
        return True

    async def is_buy_time(self, now: datetime):
        return True

    async def buy_callback(
        self,
        symbol: str,
        price: float,
        qty: float,
        now: datetime = None,
        trade_fee: float = 0.0,
    ) -> None:
        """Called by Framework, upon successful buy (could be partial)"""
        pass

    async def sell_callback(
        self,
        symbol: str,
        price: float,
        qty: float,
        now: datetime = None,
        trade_fee: float = 0.0,
    ) -> None:
        """Called by Framework, upon successful sell (could be partial)"""
        pass

    async def get_global_var(self, key):
        """implementing key-store retrival"""
        if key in self.global_var:
            return self.global_var[key]

        self.global_var[key] = (val := await KeyStore.load(key))
        return val

    async def set_global_var(self, key, value):
        """implementing key-store storing"""
        self.global_var[key] = value
        await KeyStore.save(key, value)

    @classmethod
    async def get_strategy(
        cls,
        batch_id: str,
        strategy_name: str,
        strategy_details: Dict,
        data_loader: DataLoader = None,
        ref_run_id: Optional[int] = None,
    ):
        """Internal, called by the Platform."""
        try:
            spec = importlib.util.spec_from_file_location(  # type: ignore
                "module.name", strategy_details["filename"]
            )
            custom_strategy_module = importlib.util.module_from_spec(spec)  # type: ignore
            spec.loader.exec_module(custom_strategy_module)  # type: ignore
            class_name = strategy_name

            custom_strategy = getattr(custom_strategy_module, class_name)

            if not issubclass(custom_strategy, Strategy):
                tlog(f"strategy must inherit from class {Strategy.__name__}")
                exit(0)
            strategy_details.pop("filename", None)
            s = custom_strategy(
                batch_id=batch_id,
                ref_run_id=ref_run_id,
                data_loader=data_loader,
                **strategy_details,
            )
            success = await s.create()
        except FileNotFoundError as e:
            tlog(
                f"[EXCEPTION] {e} : file not found `{strategy_details['filename']}`"
            )
            raise
        except Exception as e:
            tlog(
                f"[Error]exception of type {type(e).__name__} with args {e.args}"
            )
            traceback.print_exc()
            exit(0)
        else:
            return s if success else None



import base64
import hashlib
import hmac
import json
import os
import queue
import ssl
import time
from datetime import date, datetime, timedelta
from threading import Thread
from typing import Dict, List, Optional, Tuple

import pandas as pd
import requests
import websocket
from pytz import timezone

from liualgotrader.common import config
from liualgotrader.common.assets import get_asset_min_qty, round_asset
from liualgotrader.common.tlog import tlog
from liualgotrader.common.types import Order, QueueMapper, ThreadFlags, Trade
from liualgotrader.trading.base import Trader

utctz = timezone("UTC")


class GeminiTrader(Trader):
    gemini_api_key: Optional[str] = os.getenv("GEMINI_API_KEY")
    gemini_api_secret: Optional[str] = os.getenv("GEMINI_API_SECRET")
    base_url = "https://api.sandbox.gemini.com"
    base_websocket = "wss://api.sandbox.gemini.com"
    last_nonce = None

    def __init__(self, qm: QueueMapper = None):
        self.running_task: Optional[Thread] = None
        self.hb_task: Optional[Thread] = None
        self.send_hb = True
        self.ws = None
        self.flags: Optional[ThreadFlags] = None
        super().__init__(qm)

    @classmethod
    def _generate_request_headers(cls, payload: Dict) -> Dict:
        if not cls.gemini_api_secret or not cls.gemini_api_key:
            raise AssertionError(
                "both env variables GEMINI_API_KEY and GEMINI_API_SECRET must be set up"
            )
        t = datetime.now()
        payload_nonce = int(time.mktime(t.timetuple()) * 1000)

        if cls.last_nonce and cls.last_nonce == payload_nonce:
            payload_nonce += 1

        cls.last_nonce = payload_nonce

        payload["nonce"] = str(payload_nonce)
        encoded_payload = json.dumps(payload).encode()
        b64 = base64.b64encode(encoded_payload)
        signature = hmac.new(
            cls.gemini_api_secret.encode(), b64, hashlib.sha384
        ).hexdigest()

        return {
            "Content-Type": "text/plain",
            "Content-Length": "0",
            "X-GEMINI-APIKEY": cls.gemini_api_key,
            "X-GEMINI-PAYLOAD": b64,
            "X-GEMINI-SIGNATURE": signature,
            "Cache-Control": "no-cache",
        }

    def _generate_ws_headers(self, payload: Dict) -> Dict:
        if not self.gemini_api_secret or not self.gemini_api_key:
            raise AssertionError(
                "both env variables GEMINI_API_KEY and GEMINI_API_SECRET must be set up"
            )
        t = datetime.now()
        payload_nonce = str(int(time.mktime(t.timetuple()) * 1000))
        payload["nonce"] = payload_nonce
        encoded_payload = json.dumps(payload).encode()
        b64 = base64.b64encode(encoded_payload)
        signature = hmac.new(
            self.gemini_api_secret.encode(), b64, hashlib.sha384
        ).hexdigest()

        return {
            "X-GEMINI-APIKEY": self.gemini_api_key,
            "X-GEMINI-PAYLOAD": b64.decode(),
            "X-GEMINI-SIGNATURE": signature,
        }

    @classmethod
    def _get_order_event_type(cls, order_data: Dict) -> Order.EventType:
        return (
            Order.EventType.canceled
            if order_data["is_cancelled"] == True
            else Order.EventType.fill
            if order_data["remaining_amount"] == "0"
            else Order.EventType.partial_fill
        )

    @classmethod
    def _get_trade_event_type(cls, trade_data: Dict) -> Order.EventType:
        return (
            Order.EventType.canceled
            if trade_data["type"] == "cancelled"
            else Order.EventType.rejected
            if trade_data["type"] == "rejected"
            else Order.EventType.canceled
            if trade_data["type"] == "cancel_rejected"
            else Order.EventType.fill
            if trade_data["remaining_amount"] == "0"
            else Order.EventType.partial_fill
        )

    @classmethod
    def _get_order_side(cls, order_data: Dict) -> Order.FillSide:
        return (
            Order.FillSide.buy
            if order_data["side"] == "buy"
            else Order.FillSide.sell
        )

    @classmethod
    def _order_from_dict(cls, order_data: Dict) -> Order:
        trades = order_data.get("trades", [])
        trade_fees: float = 0.0 + sum(float(t["fee_amount"]) for t in trades)
        return Order(
            order_id=order_data["order_id"],
            symbol=order_data["symbol"].lower(),
            filled_qty=float(order_data["executed_amount"]),
            event=cls._get_order_event_type(order_data),
            price=float(order_data["price"]),
            side=cls._get_order_side(order_data),
            submitted_at=pd.Timestamp(
                ts_input=order_data["timestampms"], unit="ms", tz="UTC"
            ),
            avg_execution_price=float(order_data["avg_execution_price"]),
            remaining_amount=float(order_data["remaining_amount"]),
            trade_fees=trade_fees,
        )

    @classmethod
    def _trade_from_dict(cls, trade_dict: Dict) -> Trade:
        tlog(f"GEMINI GOING TO SEND {trade_dict}")
        return Trade(
            order_id=trade_dict["order_id"],
            symbol=trade_dict["symbol"].lower(),
            event=cls._get_trade_event_type(trade_dict),
            filled_qty=float(trade_dict["fill"]["amount"])
            if "fill" in trade_dict
            else 0.0,
            trade_fee=float(
                trade_dict["fill"]["fee"] if "fill" in trade_dict else 0.0
            )
            if "fill" in trade_dict
            else 0.0,
            filled_avg_price=float(trade_dict["avg_execution_price"] or 0.0),
            liquidity=trade_dict["fill"]["liquidity"]
            if "fill" in trade_dict
            else "",
            updated_at=pd.Timestamp(
                ts_input=trade_dict["timestampms"], unit="ms", tz="UTC"
            ),
            side=Order.FillSide[trade_dict["side"]],
        )

    async def is_fractionable(self, symbol: str) -> bool:
        return True

    def check_error(self, result: Dict):
        if result.get("result") == "error":
            raise AssertionError(
                f"[EXCEPTION] {result['reason']}:{result['message']}"
            )

    async def is_order_completed(
        self, order_id: str, external_order_id: Optional[str] = None
    ) -> Tuple[
        Order.EventType, Optional[float], Optional[float], Optional[float]
    ]:
        order = await self.get_order(order_id)
        return (
            order.event,
            order.avg_execution_price,
            order.filled_qty,
            order.trade_fees,
        )

    def get_market_schedule(
        self,
    ) -> Tuple[Optional[datetime], Optional[datetime]]:
        return datetime.now().replace(
            hour=0, minute=0, second=0, microsecond=0, tzinfo=utctz
        ), datetime.now().replace(
            hour=23, minute=59, second=59, microsecond=0, tzinfo=utctz
        )

    def get_trading_days(
        self, start_date: date, end_date: date = date.today()
    ) -> pd.DataFrame:
        return pd.DataFrame(
            index=pd.date_range(start=start_date, end=end_date)
        )

    def get_position(self, symbol: str) -> float:
        symbol = symbol.lower()
        endpoint = "/v1/balances"
        url = self.base_url + endpoint

        payload = {
            "request": endpoint,
        }
        headers = self._generate_request_headers(payload)
        response = requests.post(url, data=None, headers=headers)

        if response.status_code == 200:
            return next(
                (
                    float(b["amount"])
                    for b in response.json()
                    if b["currency"] == symbol
                ),
                0.0,
            )

        raise AssertionError(
            f"HTTP ERROR {response.status_code} {response.text}"
        )

    async def get_order(
        self, order_id: str, client_order_id: Optional[str] = None
    ) -> Order:
        endpoint = "/v1/order/status"
        url = self.base_url + endpoint

        payload = {
            "request": endpoint,
            "order_id": order_id,
            "include_trades": True,
        }
        headers = self._generate_request_headers(payload)
        response = requests.post(url, data=None, headers=headers)

        if response.status_code == 200:
            order_data = response.json()
            self.check_error(order_data)

            return self._order_from_dict(order_data)

        raise AssertionError(
            f"HTTP ERROR {response.status_code} {response.text}"
        )

    def is_market_open_today(self) -> bool:
        return True

    def get_time_market_close(self) -> Optional[timedelta]:
        return datetime.now().replace(
            hour=23, minute=59, second=59, microsecond=0, tzinfo=utctz
        ) - datetime.now().replace(tzinfo=utctz)

    async def reconnect(self):
        await self.close()
        await self.run()

    @classmethod
    def heartbeat(cls, flags: ThreadFlags):
        tlog("GEMINI HEARTBEAT thread starting")
        while flags.run:
            tlog("GEMINI HEARTBEAT")
            endpoint = "/v1/heartbeat"
            url = cls.base_url + endpoint

            payload = {
                "request": endpoint,
            }
            headers = cls._generate_request_headers(payload)
            response = requests.post(url, data=None, headers=headers)

            if response.status_code != 200:
                raise AssertionError(
                    f"HEARTHBEAT HTTP ERROR {response.status_code} {response.text}"
                )

            time.sleep(20)

        tlog("GEMINI HEARTBEAT thread terminated")

    @classmethod
    def on_message(cls, ws, msgs):
        msgs = json.loads(msgs)
        if type(msgs) != list:
            return

        for msg in msgs:
            if msg["type"] in [
                "fill",
                "cancel_rejected",
                "cancelled",
                "rejected",
            ]:
                trade = cls._trade_from_dict(msg)
                symbol = trade.symbol.lower()
                tlog(f"GEMINI TRADING UPDATE:{trade}")
                to_send = {
                    "EV": "trade_update",
                    "symbol": symbol,
                    "trade": trade.__dict__,
                }
                try:
                    if qs := cls.get_instance().queues:
                        for q in qs.get_allqueues():
                            q.put(to_send, timeout=1)
                except queue.Full as f:
                    tlog(
                        f"[EXCEPTION] : queue for {symbol} is FULL:{f}, sleeping for 2 seconds and re-trying."
                    )
                    raise

    @classmethod
    def on_error(cls, ws, error):
        tlog(f"[ERROR] GeminiTrader {error}")

    @classmethod
    def on_close(cls, ws, close_status_code, close_msg):
        tlog(f"on_close(): status={close_status_code}, close_msg={close_msg}")

    async def run(self):
        if not self.running_task:
            tlog("starting Gemini listener")
            endpoint = "/v1/order/events"
            payload = {"request": endpoint}
            headers = self._generate_ws_headers(payload)
            self.ws = websocket.WebSocketApp(
                f"{self.base_websocket}{endpoint}?eventTypeFilter=cancel_rejected&eventTypeFilter=cancelled&eventTypeFilter=rejected&eventTypeFilter=fill&eventTypeFilter=closed&heartbeat=true",
                on_message=self.on_message,
                on_error=self.on_error,
                on_close=self.on_close,
                header=headers,
            )
            self.running_task = Thread(
                target=self.ws.run_forever,
                args=(None, {"cert_reqs": ssl.CERT_NONE}),
            )
            self.flags = ThreadFlags(run=True)
            self.hb_task = Thread(target=self.heartbeat, args=(self.flags,))
            self.running_task.start()
            self.hb_task.start()

        return self.running_task

    async def close(self):
        if self.running_task and self.running_task.is_alive():
            tlog(f"close task {self.running_task}")
            self.ws.keep_running = False
            self.flags.run = False
            self.running_task.join()
            self.hb_task.join()
            tlog("task terminated")
            self.ws = None
            self.running_task = None
            self.hb_task = None
            self.flags = None

    async def get_tradeable_symbols(self) -> List[str]:
        endpoint = "/v1/symbols"
        url = self.base_url + endpoint
        response = requests.get(url)
        if response.status_code == 200:
            return response.json()

        raise AssertionError(
            f"HTTP ERROR {response.status_code} {response.text}"
        )

    async def get_shortable_symbols(self) -> List[str]:
        return []

    async def is_shortable(self, symbol) -> bool:
        return False

    async def cancel_order(self, order: Order) -> bool:
        endpoint = "/v1/order/cancel"
        url = self.base_url + endpoint

        payload = {"request": endpoint, "order_id": order.order_id}
        headers = self._generate_request_headers(payload)
        response = requests.post(url, data=None, headers=headers)
        if response.status_code == 200:
            order_status = response.json()
            self.check_error(order_status)
            return order_status

        raise AssertionError(
            f"HTTP ERROR {response.status_code} {response.text}"
        )

    async def submit_order(
        self,
        symbol: str,
        qty: float,
        side: str,
        order_type: str,
        time_in_force: str = None,
        limit_price: str = None,
        stop_price: str = None,
        client_order_id: str = None,
        extended_hours: bool = None,
        order_class: str = None,
        take_profit: dict = None,
        stop_loss: dict = None,
        trail_price: str = None,
        trail_percent: str = None,
        on_behalf_of: str = None,
    ) -> Order:
        symbol = symbol.lower()
        if order_type == "market":
            raise AssertionError(
                "GEMINI does not support market orders, use limit orders"
            )

        if qty < get_asset_min_qty(symbol):
            raise AssertionError(
                f"GEMINI requested quantity of {qty} is below minimum for {symbol}"
            )

        endpoint = "/v1/order/new"
        url = self.base_url + endpoint
        qty = round_asset(symbol, qty)
        payload = {
            "request": endpoint,
            "symbol": symbol,
            "amount": str(qty),
            "price": limit_price
            if order_type == "limit"
            else str(60000.0 * qty),
            "side": side,
            "type": "exchange limit",
            "client_order_id": client_order_id,
            "options": ["immediate-or-cancel"]
            if order_type == "market"
            else [],
        }

        headers = self._generate_request_headers(payload)
        response = requests.post(url, data=None, headers=headers)

        if response.status_code == 200:
            new_order = response.json()
            self.check_error(new_order)
            return self._order_from_dict(new_order)

        if self.flags:
            self.flags.run = False

        await self.close()

        raise AssertionError(
            f"HTTP ERROR {response.status_code} {response.text}"
        )



from datetime import datetime

import pandas as pd
from pandas import DataFrame as df
from pandas import Timestamp as ts

from liualgotrader.common import config
from liualgotrader.common.tlog import tlog


def add_daily_vwap(
    minute_data: df, debug=False, back_time=None, in_place=True
):
    if debug:
        tlog(f"before vwap {minute_data}")

    try:
        back_time_index = minute_data["close"].index.get_indexer(
            [back_time], method="nearest"
        )[0]
    except Exception as e:
        if debug:
            tlog(
                f"IndexError exception {e} in add_daily_vwap for {minute_data}"
            )
        return False

    df = minute_data if in_place else minute_data.copy()

    df["average"] = df.apply(
        lambda x: (x.close + x.high + x.low) / 3.0, axis=1
    )
    df["pv"] = df.apply(lambda x: x.average * x.volume, axis=1)
    df["apv"] = df["pv"][back_time_index:].cumsum()
    df["av"] = df["volume"][back_time_index:].cumsum()
    df["vwap"] = df["apv"] / df["av"]

    return (
        True
        if in_place
        else df[
            df.close.index.get_indexer([back_time], method="nearest")[0] :
        ].vwap
    )


def anchored_vwap(
    ohlc_data: df, start_time: datetime, debug=False
) -> pd.Series:
    return add_daily_vwap(
        ohlc_data, debug=debug, back_time=start_time, in_place=False
    )



import psutil

from liualgotrader.common import config
from liualgotrader.common.tlog import tlog


def calc_num_consumer_processes() -> int:
    if config.num_consumers > 0:
        return config.num_consumers

    num_cpu = psutil.cpu_count(False)

    if not num_cpu:
        tlog(
            "Can't automatically detect number of CPUs, use fixed configuration"
        )
        return 1

    load_pct: float = psutil.cpu_percent(interval=5)

    tlog(f"Total CPU Load:{load_pct}, num_cpu:{num_cpu}")

    return int(5.0 * (1 - load_pct / 100.0) * num_cpu)



import asyncio
import json
import queue
import time as time_action
from datetime import date, datetime, time, timedelta
from threading import Thread
from typing import Dict, List, Optional, Tuple

import pandas as pd
import pytz
import requests
import websocket

from liualgotrader.common import config
from liualgotrader.common.assets import get_asset_min_qty, round_asset
from liualgotrader.common.tlog import tlog
from liualgotrader.common.types import Order, QueueMapper, ThreadFlags, Trade
from liualgotrader.trading.base import Trader

NY = "America/New_York"
nytz = pytz.timezone(NY)


class TradierTrader(Trader):
    order_symbol: Dict[int, str] = {}
    order_side: Dict[int, Order.FillSide] = {}

    def __init__(self, qm: QueueMapper = None):
        self.running_task: Optional[Thread] = None
        self.ws: Optional[websocket] = None
        self.ws_session_id: Optional[str] = None
        super().__init__(qm)

    def _get(self, url: str, params: Optional[Dict] = None):
        if params is None:
            params = {}
        r = requests.get(
            url,
            params=params,
            headers={
                "Authorization": f"Bearer {config.tradier_access_token}",
                "Accept": "application/json",
            },
        )
        if r.status_code in (429, 502):
            tlog(f"{r.url} return {r.status_code}, waiting and re-trying")
            time_action.sleep(10)
            return self._get(url, params)
        elif r.status_code != 200:
            raise ValueError(f"HTTP ERROR {r.url} {r.status_code} {r.text}")

        return r

    def _post(self, url: str, params: Optional[Dict] = None):
        if params is None:
            params = {}
        r = requests.post(
            url,
            params=params,
            headers={
                "Authorization": f"Bearer {config.tradier_access_token}",
                "Accept": "application/json",
            },
        )
        if r.status_code in (429, 502):
            tlog(f"{r.url} return {r.status_code}, waiting and re-trying")
            time_action.sleep(10)
            return self._post(url, params)
        elif r.status_code != 200:
            raise ValueError(f"HTTP ERROR {r.url} {r.status_code} {r.text}")

        return r

    def _delete(self, url: str, params: Optional[Dict] = None):
        if params is None:
            params = {}
        r = requests.delete(
            url,
            params=params,
            headers={
                "Authorization": f"Bearer {config.tradier_access_token}",
                "Accept": "application/json",
            },
        )
        if r.status_code in (429, 502):
            tlog(f"{r.url} return {r.status_code}, waiting and re-trying")
            time_action.sleep(10)
            return self._delete(url, params)
        elif r.status_code != 200:
            raise ValueError(f"HTTP ERROR {r.url} {r.status_code} {r.text}")

        return r

    async def is_fractionable(self, symbol: str) -> bool:
        return False

    async def is_order_completed(
        self, order_id: str, external_order_id: Optional[str] = None
    ) -> Tuple[Order.EventType, float, float, float]:
        order = await self.get_order(order_id, external_order_id)
        return (
            order.event,
            float(order.avg_execution_price or 0.0),
            float(order.filled_qty or 0.0),
            0.0,
        )

    def get_market_schedule(
        self,
    ) -> Tuple[Optional[datetime], Optional[datetime]]:
        url = f"{config.tradier_base_url}markets/clock"
        response = self._get(
            url,
        )
        data = response.json()

        last_open_date = datetime.strptime(
            data["clock"]["date"], "%Y-%m-%d"
        ).date()
        if last_open_date == date.today():
            start: Optional[datetime] = nytz.localize(
                datetime.combine(
                    date=last_open_date,
                    time=time(hour=7, minute=0, second=0, microsecond=0),
                )
            )

            end: Optional[datetime] = nytz.localize(
                datetime.combine(
                    date=last_open_date,
                    time=time(
                        hour=19, minute=55, second=59, microsecond=999999
                    ),
                )
            )
        else:
            start = end = None

        return start, end

    def get_trading_days(
        self, start_date: date, end_date: date = date.today()
    ) -> pd.DataFrame:
        url = f"{config.tradier_base_url}markets/calendar"
        d = start_date
        df = pd.DataFrame()

        while True:
            response = self._get(
                url, params={"month": d.month, "year": d.year}
            )

            data = response.json()["calendar"]["days"]["day"]
            dict_data: Dict = {
                element["date"]: element
                for element in data
                if element["status"] == "open"
            }
            df1 = pd.DataFrame.from_dict(dict_data, orient="index")
            df1["close"] = df1.apply(
                lambda x: x.get("postmarket", x["open"])["end"], axis=1
            )
            df1["open"] = df1.apply(lambda x: x["premarket"]["start"], axis=1)

            df1.index = pd.to_datetime(df1.index)
            df1 = df1[["open", "close"]].loc[
                (datetime.combine(start_date, time.min) <= df1.index)
                & (df1.index <= datetime.combine(end_date, time.min))
            ]
            df = pd.concat([df, df1])
            if d.month == end_date.month and d.year == end_date.year:
                break

            try:
                d = date(month=d.month + 1, day=d.day, year=d.year)
            except ValueError:
                d = date(month=d.month + 1, day=d.day - 1, year=d.year)

        return df

    def get_position(self, symbol: str) -> float:
        url = f"{config.tradier_base_url}accounts/{config.tradier_account_number}/positions"
        response = self._get(
            url,
        )
        data = response.json()
        if data["positions"] == "null":
            return 0.0

        data = data["positions"]["position"]
        if isinstance(data, dict):
            return data["quantity"] if data["symbol"] == symbol else 0.0

        return next(
            (
                position["quantity"]
                for position in data
                if position["symbol"] == symbol
            ),
            0.0,
        )

    @classmethod
    def _get_event(cls, order_event: str) -> Order.EventType:
        return (
            Order.EventType.fill
            if order_event == "filled"
            else Order.EventType.partial_fill
            if order_event == "partially_filled"
            else Order.EventType.canceled
            if order_event == "canceled"
            else Order.EventType.rejected
            if order_event == "rejected"
            else Order.EventType.pending
            if order_event == "pending"
            else Order.EventType.error
            if order_event == "error"
            else Order.EventType.open
            if order_event == "open"
            else Order.EventType.other
        )

    async def get_order(
        self, order_id: str, client_order_id: Optional[str] = None
    ) -> Order:
        url = f"{config.tradier_base_url}accounts/{config.tradier_account_number}/orders/{order_id}"
        response = self._get(
            url,
        )
        data = response.json()
        return Order(
            order_id=data["order"]["id"],
            symbol=data["order"]["symbol"],
            event=self._get_event(data["order"]["status"]),
            submitted_at=data["order"]["create_date"],
            price=data["order"]["last_fill_price"],
            trade_fees=0.0,
            filled_qty=data["order"]["exec_quantity"],
            side=self._get_event_side(data["order"]["side"]),
            remaining_amount=data["order"]["remaining_quantity"],
            avg_execution_price=data["order"]["avg_fill_price"],
            external_account_id=data["order"]["tag"]
            if "tag" in data["order"]
            else None,
        )

    def is_market_open_today(self) -> bool:
        url = f"{config.tradier_base_url}markets/calendar"
        response = self._get(
            url,
        )
        data = response.json()
        today = str(datetime.now(nytz).date())
        return next(
            (
                day["status"] == "open"
                for day in data["calendar"]["days"]["day"]
                if day["date"] == today
            ),
            False,
        )

    def get_time_market_close(self) -> Optional[timedelta]:
        url = f"{config.tradier_base_url}markets/calendar"
        response = self._get(
            url,
        )
        data = response.json()
        today = str(datetime.now(nytz).date())
        return next(
            (
                day.get("postmarket", day["open"])["end"]
                for day in data["calendar"]["days"]["day"]
                if day["date"] == today and "open" in day
            ),
            None,
        )

    async def get_tradeable_symbols(self) -> List[str]:
        return await self.get_shortable_symbols()

    async def get_shortable_symbols(self) -> List[str]:
        url = f"{config.tradier_base_url}markets/etb"
        response = self._get(
            url,
        )
        data = response.json()
        return [
            security["symbol"]
            for security in data["securities"]["security"]
            if security["type"] == "stock"
        ]

    async def is_shortable(self, symbol) -> bool:
        return symbol.upper() in await self.get_shortable_symbols()

    async def cancel_order(self, order: Order) -> bool:
        url = f"{config.tradier_base_url}accounts/{config.tradier_account_number}/orders/{order.order_id}"

        response = self._delete(
            url,
        )
        data = response.json()
        return data["order"]["status"] == "ok"

    @classmethod
    def _get_event_side(cls, side: str) -> Order.FillSide:
        if side in {"buy", "buy_to_cover"}:
            return Order.FillSide.buy

        return Order.FillSide.sell

    async def subscribe_order_status_update(self) -> None:
        if (
            self.ws is not None
            and self.ws_session_id
            and self.running_task
            and self.running_task.is_alive()
        ):
            payload = {
                "events": ["order"],
                "sessionid": self.ws_session_id,
                "excludeAccounts": [],
            }
            self.ws.send(json.dumps(payload))
            tlog(
                f"subscribed for order update on session {self.ws_session_id}"
            )

    async def submit_order(
        self,
        symbol: str,
        qty: float,
        side: str,
        order_type: str,
        time_in_force: str = None,
        limit_price: str = None,
        stop_price: str = None,
        client_order_id: str = None,
        extended_hours: bool = None,
        order_class: str = None,
        take_profit: dict = None,
        stop_loss: dict = None,
        trail_price: str = None,
        trail_percent: str = None,
        on_behalf_of: str = None,
    ) -> Order:
        url = f"{config.tradier_base_url}accounts/{config.tradier_account_number}/orders"

        params: Dict = {
            "class": "equity",
            "symbol": symbol,
            "side": side,
            "quantity": str(int(qty)),
            "duration": "day",
            "type": order_type,
        }

        if limit_price:
            params["price"] = str(round(float(limit_price), 2))
        if stop_price:
            params["stop"] = str(round(float(stop_price), 2))
        if on_behalf_of:
            params["tag"] = on_behalf_of

        response = self._post(
            url,
            params,
        )
        data = response.json()

        if "errors" in data:
            raise ValueError(f"order failed with {data['errors']}")

        event_side = self._get_event_side(side)
        o = Order(
            symbol=symbol,
            order_id=data["order"]["id"],
            event=Order.EventType.pending,
            side=event_side,
            submitted_at=datetime.now(nytz),
            remaining_amount=qty,
        )

        if on_behalf_of:
            o.external_account_id = on_behalf_of
        if limit_price:
            o.price = float(limit_price)

        TradierTrader.order_symbol[int(o.order_id)] = symbol
        TradierTrader.order_side[int(o.order_id)] = event_side
        return o

    async def reconnect(self):
        await self.close()
        await self.run()

    async def run(self) -> Optional[asyncio.Task]:
        if not self.running_task:
            tlog("starting Tradier Account listener")

            url = f"{config.tradier_base_url}accounts/events/session"

            response = self._post(
                url,
            )
            data = response.json()

            ws_url = data["stream"]["url"]
            self.ws_session_id = data["stream"]["sessionid"]

            tlog(
                f"starting WebSocketApp on {ws_url} with session-id {self.ws_session_id}"
            )
            self.ws = websocket.WebSocketApp(
                ws_url,
                on_message=self.on_message,
                on_error=self.on_error,
                on_close=self.on_close,
            )
            self.running_task = Thread(
                target=self.ws.run_forever,
            )
            self.running_task.start()
            await asyncio.sleep(1.0)
            await self.subscribe_order_status_update()

        return self.running_task  # type: ignore

    async def close(self):
        if self.running_task and self.running_task.is_alive():
            tlog(f"close task {self.running_task}")
            self.ws.keep_running = False
            self.running_task.join()
            tlog("task terminated")
            self.ws = None
            self.running_task = None
            self.ws_session_id = None

    @classmethod
    def _trade_from_dict(cls, trade_dict: Dict) -> Trade:
        return Trade(
            order_id=trade_dict["id"],
            symbol=TradierTrader.order_symbol[trade_dict["id"]],
            event=cls._get_event(trade_dict["status"]),
            side=TradierTrader.order_side[trade_dict["id"]],
            filled_qty=float(trade_dict.get("last_fill_quantity", 0.0)),
            trade_fee=0.0,
            filled_avg_price=float(trade_dict.get("last_fill_price", 0.0)),
            liquidity="",
            updated_at=trade_dict["create_date"],
        )

    @classmethod
    def on_message(cls, ws, msgs):
        msg = json.loads(msgs)

        if msg["event"] != "order":
            return

        if (
            msg["status"]
            in {
                "filled",
                "cancel_rejected",
                "cancelled",
                "rejected",
            }
            and msg["id"] in TradierTrader.order_symbol
        ):
            trade = cls._trade_from_dict(msg)
            symbol = trade.symbol.lower()
            tlog(f"TRADIER TRADING UPDATE:{trade}")
            to_send = {
                "EV": "trade_update",
                "symbol": symbol,
                "trade": trade.__dict__,
            }
            try:
                if qs := cls.get_instance().queues:
                    for q in qs.get_allqueues():
                        q.put(to_send, timeout=1)
            except queue.Full as f:
                tlog(
                    f"[EXCEPTION] queue for {symbol} is FULL:{f}, sleeping for 2 seconds and re-trying."
                )
                raise

    @classmethod
    def on_error(cls, ws, error):
        tlog(f"[ERROR] TradierTrader {error}")

    @classmethod
    def on_close(cls, ws, close_status_code, close_msg):
        tlog(
            f"on_close(): TradierTrader status={close_status_code}, close_msg={close_msg}"
        )



import math
from abc import ABCMeta, abstractmethod
from datetime import date, datetime
from typing import Awaitable, Callable, Dict, List, Optional

import pandas as pd

from liualgotrader.common.types import TimeScale


class DataAPI(metaclass=ABCMeta):
    def __init__(self, ws_uri: str, ws_messages_handler: Awaitable):
        self.ws_uri = ws_uri
        self.ws_msgs_handler = ws_messages_handler

    @abstractmethod
    def get_symbol_data(
        self,
        symbol: str,
        start: date,
        end: date = date.today(),
        scale: TimeScale = TimeScale.minute,
    ) -> pd.DataFrame:
        ...

    @abstractmethod
    async def get_market_snapshot(
        self, filter_func: Optional[Callable]
    ) -> List[Dict]:
        ...

    @abstractmethod
    def get_symbols(self) -> List[str]:
        ...

    @abstractmethod
    def get_symbols_data(
        self,
        symbols: List[str],
        start: date,
        end: date = date.today(),
        scale: TimeScale = TimeScale.minute,
    ) -> Dict[str, pd.DataFrame]:
        ...

    @abstractmethod
    def get_last_trading(self, symbol: str) -> datetime:
        ...

    @abstractmethod
    def get_trading_day(
        self, symbol: str, now: datetime, offset: int
    ) -> datetime:
        ...

    @abstractmethod
    def trading_days_slice(self, symbol: str, slice) -> slice:
        ...

    @abstractmethod
    def num_trading_minutes(self, symbol: str, start: date, end: date) -> int:
        ...

    @abstractmethod
    def num_trading_days(self, symbol: str, start: date, end: date) -> int:
        ...

    @abstractmethod
    def get_max_data_points_per_load(self) -> int:
        ...

    def data_concurrency_ranges(
        self, symbol: str, start: date, end: date, scale: TimeScale
    ) -> List[Optional[pd.DatetimeIndex]]:
        scale_factor_minutes = self.num_trading_minutes(symbol, start, end)
        data_points = (
            scale_factor_minutes / 60
            if scale == TimeScale.day
            else scale_factor_minutes
        )
        total_data_points = (
            self.num_trading_days(symbol, start, end) * data_points
        )
        periods = math.ceil(
            total_data_points / self.get_max_data_points_per_load()
        )
        return pd.date_range(start, end, periods=periods + 1)



import io
from datetime import date, datetime, timedelta
from typing import Dict, List, Optional

import pandas as pd
import pytz
import requests
from finnhub import Client

from liualgotrader.common import config
from liualgotrader.common.types import TimeScale
from liualgotrader.data import static
from liualgotrader.data.data_base import DataAPI

NY = "America/New_York"
nytz = pytz.timezone(NY)


def check_auth(f):
    def wrapper(*args):
        if not args[0].finnhub_rest_client:
            raise AssertionError("Must call w/ authenticated Finnhub client")
        return f(*args)

    return wrapper


class FinnhubData(DataAPI):
    def __init__(self):
        self.finnhub_rest_client = Client(api_key=config.finnhub_api_key)
        if not self.finnhub_rest_client:
            raise AssertionError("Failed to authenticate Finnhub  client")

        s = requests.get(static.finnhub_exchanges_url).content
        self.stock_exchanges: pd.DataFrame = pd.read_csv(
            io.StringIO(s.decode("utf-8"))
        )

    @check_auth
    def get_symbols(
        self,
        country: str = "US",
    ) -> List[Dict]:
        if country not in self.stock_exchanges.code.to_list():
            raise AssertionError(
                f"country code {country} not supported, valid values are {self.stock_exchanges.code.to_list()}"
            )
        return self.finnhub_rest_client.stock_symbols(exchange=country)

    # @check_auth
    def get_symbol_data(
        self,
        symbol: str,
        start: date,
        end: date = date.today(),
        scale: TimeScale = TimeScale.minute,
    ) -> pd.DataFrame:
        _start = int(datetime.combine(start, datetime.min.time()).timestamp())
        _end = int(
            datetime.combine(
                end if scale == TimeScale.day else end + timedelta(days=1),
                datetime.min.time(),
            ).timestamp()
        )
        t: Optional[str] = (
            "1"
            if scale == TimeScale.minute
            else "D"
            if scale == TimeScale.day
            else None
        )

        if not t:
            raise AssertionError(
                f"timescale {scale} not support in Finnhub implementation"
            )
        data = pd.DataFrame(
            self.finnhub_rest_client.stock_candles(
                symbol,
                t,
                _start,
                _end,
            )
        )
        data.t = pd.to_datetime(data.t, unit="s")
        data = data.set_index(data.t)
        data = data.tz_localize("America/New_York", ambiguous="infer")
        data = data[data.s == "ok"]
        data = data.drop(columns=["s", "t"])
        data.rename(
            columns={
                "o": "open",
                "c": "close",
                "h": "high",
                "l": "low",
                "v": "volume",
            },
            inplace=True,
        )

        if data.empty:
            raise ValueError(
                f"[ERROR] {symbol} has no data for {_start} to {_end} w {scale.name}"
            )

        return data

    def get_symbols_data(
        self,
        symbols: List[str],
        start: date,
        end: date = date.today(),
        scale: TimeScale = TimeScale.minute,
    ) -> Dict[str, pd.DataFrame]:
        raise NotImplementedError("get_symbols_data")

    def get_last_trading(self, symbol: str) -> datetime:
        raise NotImplementedError("get_last_trading")

    def get_trading_day(
        self, symbol: str, now: datetime, offset: int
    ) -> datetime:
        raise NotImplementedError("get_trading_day")

    async def get_market_snapshot(self, filter_func) -> List[Dict]:
        raise NotImplementedError

    def trading_days_slice(self, symbol: str, slice) -> slice:
        raise NotImplementedError("trading_days_slice")

    def num_trading_minutes(self, symbol: str, start: date, end: date) -> int:
        raise NotImplementedError("num_trading_minutes")

    def num_trading_days(self, symbol: str, start: date, end: date) -> int:
        raise NotImplementedError("num_trading_days")

    def get_max_data_points_per_load(self) -> int:
        raise NotImplementedError("get_max_data_points_per_load")



import time as time_action
from datetime import date, datetime, time
from typing import Callable, Dict, List, Optional

import pandas as pd
import pandas_market_calendars
import pytz
import requests

from liualgotrader.common import config
from liualgotrader.common.tlog import tlog
from liualgotrader.common.types import TimeScale
from liualgotrader.data.data_base import DataAPI

NY = "America/New_York"
nytz = pytz.timezone(NY)


class TradierData(DataAPI):
    datapoints_per_request = 500
    max_trades_per_minute = 10

    def __init__(self):
        ...

    def _get(self, url: str, params: Dict = None):
        if params is None:
            params = {}
        r = requests.get(
            url,
            params=params,
            headers={
                "Authorization": f"Bearer {config.tradier_access_token}",
                "Accept": "application/json",
            },
        )
        if r.status_code in (429, 502):
            tlog(f"{url} return {r.status_code}, waiting and re-trying")
            time_action.sleep(10)
            return self._get(url, params)

        return r

    def get_symbol_data(
        self,
        symbol: str,
        start: date,
        end: date = date.today(),
        scale: TimeScale = TimeScale.minute,
    ) -> pd.DataFrame:
        if scale == TimeScale.day:
            url = f"{config.tradier_base_url}markets/history"
            interval = "daily"
            s = str(start)
            e = str(end)
        elif scale == TimeScale.minute:
            url = f"{config.tradier_base_url}markets/timesales"
            interval = "1min"
            s = datetime.combine(start, datetime.min.time()).strftime(
                "%Y-%m-%d %H:%M"
            )
            e = datetime.combine(end, datetime.max.time()).strftime(
                "%Y-%m-%d %H:%M"
            )
        else:
            raise NotImplementedError(f"scale {scale} not implemented yet")

        response = self._get(
            url,
            params={
                "symbol": symbol,
                "interval": interval,
                "start": s,
                "end": e,
            },
        )

        if response.status_code != 200:
            raise ValueError(
                f"HTTP ERROR {response.status_code} {response.text}"
            )

        df: pd.DataFrame = pd.DataFrame()
        if scale == TimeScale.day:
            data = response.json()["history"]["day"]
            df = pd.DataFrame(data=data)

            df.date = pd.to_datetime(df.date).dt.tz_localize("EST")
            df.set_index(
                "date", drop=True, inplace=True, verify_integrity=True
            )
            df["count"] = 0
            df["average"] = 0.0
            df["vwap"] = 0.0
        elif scale == TimeScale.minute:
            data = response.json()

            try:
                if data and data["series"]:
                    data = data["series"]["data"]
                    df = pd.DataFrame.from_records(data)
                    df.time = pd.to_datetime(df.time).dt.tz_localize("EST")
                    df.set_index(
                        "time", drop=True, inplace=True, verify_integrity=True
                    )
                    df.drop(["timestamp"], axis=1, inplace=True)
                    df["count"] = 0
                    df = df.rename(columns={"price": "average"}).between_time(
                        "04:00", "19:59"
                    )
            except Exception as exc:
                raise ValueError(f"data={data}") from exc

        return df

    async def get_market_snapshot(
        self, filter_func: Optional[Callable]
    ) -> List[Dict]:
        raise NotImplementedError("get_market_snapshot")

    def get_symbols(self) -> List[str]:
        raise NotImplementedError("get_symbols")

    def get_symbols_data(
        self,
        symbols: List[str],
        start: date,
        end: date = date.today(),
        scale: TimeScale = TimeScale.minute,
    ) -> Dict[str, pd.DataFrame]:
        ...

    def _get_previous_month_last_trading(
        self, month: int, year: int
    ) -> datetime:
        url = f"{config.tradier_base_url}markets/calendar"
        response = self._get(url, params={"month": month})
        data = response.json()
        return self._get_previous_last_trading(
            data, len(data["calendar"]["days"]["day"]) - 1, month, year
        )

    def _get_previous_last_trading(
        self, data: Dict, index: int, month: int, year: int
    ) -> datetime:
        if index > 0:
            if data["calendar"]["days"]["day"][index - 1]["status"] == "open":
                date_str = data["calendar"]["days"]["day"][index - 1]["date"]
                day = datetime.strptime(date_str, "%Y-%m-%d").date()
                return nytz.localize(datetime.combine(day, time(hour=20)))
            return self._get_previous_last_trading(
                data, index - 1, month, year
            )
        return (
            self._get_previous_month_last_trading(month=month - 1, year=year)
            if month > 0
            else self._get_previous_month_last_trading(month=12, year=year - 1)
        )

    def get_last_trading(self, symbol: str) -> datetime:
        url = f"{config.tradier_base_url}markets/calendar"
        response = self._get(
            url,
        )
        data = response.json()
        today = datetime.now(nytz).date()
        for i, day in enumerate(data["calendar"]["days"]["day"]):
            if day["date"] == str(today):
                if day["status"] == "open":
                    return datetime.now(tz=nytz)

                return self._get_previous_last_trading(
                    data, i, today.month, today.year
                )
            i += 1

        raise AssertionError(
            f"Could not find {today} in current trading month"
        )

    def get_trading_holidays(self) -> List[str]:
        nyse = pandas_market_calendars.get_calendar("NYSE")
        return nyse.holidays().holidays

    def get_trading_day(
        self, symbol: str, now: datetime, offset: int
    ) -> datetime:
        cbd_offset = pd.tseries.offsets.CustomBusinessDay(
            n=offset, holidays=self.get_trading_holidays()
        )

        return nytz.localize(now + cbd_offset)

    def trading_days_slice(self, symbol: str, time_slice) -> slice:
        raise NotImplementedError("trading_days_slice")

    def num_trading_minutes(self, symbol: str, start: date, end: date) -> int:
        raise NotImplementedError("num_trading_minutes")

    def num_trading_days(self, symbol: str, start: date, end: date) -> int:
        raise NotImplementedError("num_trading_days")

    def get_max_data_points_per_load(self) -> int:
        raise NotImplementedError("get_max_data_points_per_load")



import math
from datetime import datetime, timedelta
from enum import Enum
from typing import Tuple

import numpy as np
import pandas as pd
import pytz
from scipy.stats import linregress

from liualgotrader.common.data_loader import DataLoader  # type: ignore

est = pytz.timezone("US/Eastern")


class SeriesTrendType(Enum):
    UNKNOWN = 0
    SHARP_DOWN = 1
    DOWN = 5
    UP = 15
    SHARP_UP = 20


class VolatilityClassificationType(Enum):
    UNKNOWN = 0
    LOW = 1
    MEDIUM = 5
    HIGH = 10


def get_series_trend(series: pd.Series) -> Tuple[float, SeriesTrendType]:
    if len(series) < 4:
        return 0, SeriesTrendType.UNKNOWN

    try:
        np.seterr(all="raise")
        slope, _, _, _, _ = linregress(np.arange(len(series)), series)
        slope = round(slope, 3)
    except FloatingPointError:
        return math.inf, SeriesTrendType.UNKNOWN

    if 0 < slope <= 1.0:
        t = SeriesTrendType.UP
    elif slope > 1.0:
        t = SeriesTrendType.SHARP_UP
    elif -1.0 <= slope < 0:
        t = SeriesTrendType.DOWN
    else:
        t = SeriesTrendType.SHARP_DOWN

    return slope, t


def volatility(data_loader: DataLoader, symbol: str, now: datetime) -> float:
    return (
        data_loader[symbol]
        .close[now - timedelta(days=30) : now]  # type: ignore
        .pct_change()
        .rolling(20)
        .std()
        .iloc[-1]
    )
